(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program.
Require Import vgtac basis lang_src_lsi lang_mid.
Require Import model model_properties gw.
Set Implicit Arguments. 


(** * Translation from src to mid via CPS. *)


Definition next := S.

Definition Ret (f: lang_mid.exp -> lang_mid.exp) (k: nat * lang_mid.exp) : nat * lang_mid.exp :=
  let (mx, e) := k in (mx, f e).


(** ** CPS translation of expressions.

    The translation makes sure that in the resulting expresssion no
    variable is used twice.  To do so, it takes as input an unused
    variable [mx] and assumes that any larger variable is also unused
    (recall that variables are natural numbers).  Whenever a bound
    variable from the source expression is about to end up in the mid
    expression, it is first renamed to an unused one.  Furthermore,
    whenever the translation introduces a new bound variable, this
    variable will be taken from the unused ones.  To maintain the set
    of unused ones, an updated number is returned along with the
    converted expression.  Note that the meta-level continuations used
    in the translation also take and return this number.
*)
Fixpoint cps (e: lang_src.exp) (mx: nat) (ren: termvar -> itermvar)
             (k: nat -> itermvar -> nat * lang_mid.exp) : nat * lang_mid.exp :=
  match e with
    | Eloc _ => (mx, IEunit) 

    | Elbl l => k mx (inl l)
    | Evar x => k mx (ren x)

    | Eunit  => Ret (IElet IEunit (inl mx)) (k (next mx) (inr mx))
    | Enat n => Ret (IElet (IEnat n) (inl mx)) (k (next mx) (inr mx))
    | Einput => Ret (IEinput mx) (k (next mx) (inr mx))

    | Eroll e   => cps e mx ren k
    | Eunroll e => cps e mx ren k
    | Epack e   => cps e mx ren k

    | Efst  e => cps e mx ren (fun mx y => Ret (IElet (IEfst y) (inl mx)) (k (next mx) (inr mx)))
    | Esnd  e => cps e mx ren (fun mx y => Ret (IElet (IEsnd y) (inl mx)) (k (next mx) (inr mx)))
    | Einl  e => cps e mx ren (fun mx y => Ret (IElet (IEinl y) (inl mx)) (k (next mx) (inr mx)))
    | Einr  e => cps e mx ren (fun mx y => Ret (IElet (IEinr y) (inl mx)) (k (next mx) (inr mx)))
    | Enew  e => cps e mx ren (fun mx y => Ret (IEnew y mx) (k (next mx) (inr mx)))
    | Eread e => cps e mx ren (fun mx y => Ret (IEread y mx) (k (next mx) (inr mx)))

    | Eoutput e =>
        cps e mx ren (fun mx y =>
        Ret (fun re => IElet IEunit (inl mx) (IEoutput y re)) (k (next mx) (inr mx)))

    | Epair ea eb =>
        cps ea mx ren (fun mx ya =>
        cps eb mx ren (fun mx yb =>
        Ret (IElet (IEpair ya yb) (inl mx)) (k (next mx) (inr mx))))
    | Ebop op ea eb =>
        cps ea mx ren (fun mx ya =>
        cps eb mx ren (fun mx yb =>
        Ret (IElet (IEbop op (inl ya) (inl yb)) (inl mx)) (k (next mx) (inr mx))))
    | Ewrite ea eb =>
        cps ea mx ren (fun mx ya =>
        cps eb mx ren (fun mx yb => 
        Ret (fun re => IElet IEunit (inl mx) (IEwrite ya yb re)) (k (next mx) (inr mx))))

    | Eifnz e ea eb =>
        cps e mx ren (fun mx y =>
        let (mx', ea') := cps ea mx ren k in
        Ret (IEifnz y ea') (cps eb mx' ren k))

    | Eapp ea eb =>
        cps ea mx ren (fun mx ya =>
        cps eb mx ren (fun mx yb =>
                           Ret (fun re => IElet (IEcont (next mx) re) (inr mx) (IEapp ya yb mx))
            (k (next (next mx)) (inr (next mx)))))
    | Einst e =>
        cps e mx ren (fun mx y => 
        Ret (fun re => IElet (IEcont (next mx) re) (inr mx) (IEinst y mx))
            (k (next (next mx)) (inr (next mx))))

    | Ecase e x ea eb =>
        cps e mx ren (fun mx y =>
        let (mx', ea') := cps ea (next mx) [inr mx//x]>>ren k in
        Ret (IEcase y mx ea') (cps eb mx' [inr mx//x]>>ren k))

    | Eunpack x ea eb =>
        cps ea mx ren (fun mx y =>
        cps eb mx [y//x]>>ren k)
        (** y might be a label *)

    | Efix f x e =>
        let xf := next mx in
        let xv := next xf in
        let xk := next xv in
        let ren' := [inr xv//x]>>[inr xf//f]>>ren in
        let (mx', e') := cps e (next xk) ren' (fun mx y => (mx, IEret xk y)) in
        Ret (IElet (IEfix xf xv xk e') (inl mx)) (k mx' (inr mx))

    | Egen e =>
        let xk := next mx in
        let (mx', e') := cps e (next xk) ren (fun mx' y => (mx', IEret xk y)) in
        Ret (IElet (IEgen xk e') (inl mx)) (k mx' (inr mx))
            
    | Erefeq ea eb =>
        cps ea mx ren (fun mx ya =>
        cps eb mx ren (fun mx yb =>
        Ret (IElet (IErefeq ya yb) (inl mx)) (k (next mx) (inr mx))))
  end.


(** ** CPS translation of top-level functions. *)
Definition tlcps (mx: termvar) v : termvar * lang_mid.exp :=
  match v with
    | Efix f x e =>
        let xf := mx in
        let xx := next mx in
        let xk := next (next mx) in
        let ren := [inr xx//x]>>[inr xf//f]>>inr in
        let (mx', e') := cps e (next xk) ren (fun mx y => (mx, IEret xk y)) in
        (next mx', IEfix xf xx xk e')
    | Egen e =>
        let (mx', e') := cps e (next mx) inr (fun mx' y => (mx', IEret mx y)) in
        (next mx', IEgen mx e')
    | _ =>                      
        (mx, IEunit)
  end.


(** ** CPS translation of modules. *)
Fixpoint mcps_ mx (m: lang_src.module) : lang_mid.module :=
  match m with
    | [] => []
    | (l, v) :: m' =>
      let (mx', iv) := tlcps mx v in
      (l, iv) :: mcps_ mx' m'
  end.

Definition mcps (m: lang_src.module) : module :=
  mcps_ 0 m.
