(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

(** Abstract language specification *)

Require Import Arith Bool List Program Relations.
Require Import vgtac basis lang_common.
Set Implicit Arguments.


Reserved Notation "x # y" (at level 60, right associativity).

(** Main definition. *)

Structure LangSpec :=
  (** run-time values *)
  { val : Type
  (** continuations *)
  ; cont : Type
  (** logical machine configurations *)
  ; conf : Type
  (** physical machine configurations *)
  ; mach : Type
  (** modules *)
  ; module : Type
  (** abstract load addresses *)
  ; anchor : Type
  (** getting the runtime value of a particular module component *)
  ; vload : module -> anchor -> list (label * val) -> label -> val -> Prop
  (** loading a module into memory *)
  ; cload : module -> anchor -> list (label * val) -> conf -> conf -> Prop
  (** merging two configurations *)
  ; cplus : conf -> conf -> conf where "x # y" := (cplus x y)
  (** empty configuration *)
  ; cemp : conf
  (** machine execution as a labelled transition relation *)
  ; step : event -> mach -> mach -> Prop
  (** machines realizing a configuration *)
  ; real : conf -> mach -> Prop
  (** configurations that don't contain control information (i.e. only heap/stack) *)
  ; extra : conf -> Prop
  (** configurations that contain exactly control information *)
  
  
  ; core : conf -> Prop
  (** configurations that result from a run-time error *)
  ; error (m: mach) := forall c, ~ real c m
  (** finished machines *)
  ; halted : mach -> Prop
  ; halted_def: forall m, halted m <-> forall evt m', ~ step evt m m'
  (** merging a machine with a configuration (derived) *)
  ; mcplus (m: mach) (c: conf) (mc: mach) : Prop :=
      exists c', <<RLm: real c' m>> /\ <<RL: real (cplus c' c) mc>>
  (** [cplus] and [emp] form a commutative monoid *)
  ; cplus_comm: forall a b, a # b = b # a
  ; cplus_assoc: forall a b c, a # b # c = (a # b) # c
  ; cplus_emp: forall a, cemp # a = a
  (** an error cannot be undone *)
  ; step_error: forall evt m m' (ST: step evt m m') (RL: ~ error m'), ~ error m
  (** an error can only be produced silently (this is not needed in
      the metatheory but without it you won't be able to prove some simple examples) *)
  ; step_etau_error: forall evt m m' (NE: ~ error m) (ST: step evt m m') (E: error m'),
                       evt = etau
  (** the core of a realizable configuration is unique *)
  ; core_uniq: forall coa cob (COa: core coa) (COb: core cob)
                      ca cb m (RLa: real (coa # ca) m) (RLb: real (cob # cb) m),
                 coa = cob /\ ca = cb 
  
  (** empty configuration is [extra] *)
  ; extra_emp: extra cemp
  (** [cplus] preserves [extra] *)
  ; extra_plus: forall c c' (XTR: extra c) (XTR': extra c'), extra (c # c')
  (** [real] does not depend on [extra] configurations *)
  ; extra_real: forall c, exists m,
                  forall cf mm (RL: real (c # cf) mm) (XTR: extra cf),
                    real c m
     
  (** [halted] does not depend on [extra] configurations *)
  ; extra_halted: forall m cf mm (PL: mcplus m cf mm) (XTR: extra cf) (HA: halted mm),
                    halted m
  }.
Arguments cplus [l] _ _.
Arguments cemp [l].
Arguments core [l] _.
Arguments extra [l] _.
Arguments cplus_comm [l] _ _.
Arguments cplus_assoc [l] _ _ _.

(** a machine that is neither erroneous nor halted can step further *)
Lemma progress_if_not_halted L: forall m (NT: ~ L.(halted) m), 
  exists evt m', L.(step) evt m m'.
Proof.
  ii; apply not_all_not_ex.
  ii; apply NT.
  apply halted_def.
  ii; eapply H; eauto.
Qed.

(** a halted machine cannot step *)
Lemma halted_step L: forall m (HA: L.(halted) m), 
  forall evt m', ~ L.(step) evt m m'.
Proof.
  i; eapply halted_def; eauto.
Qed.

Infix "#" := cplus (at level 60, right associativity).


Lemma cplus_swap: forall L (a b c: L.(conf)),
  a # b # c = b # a # c.
Proof.
  i; rewrite cplus_assoc, (cplus_comm a), <-cplus_assoc; auto.
Qed.


(** Simplify combinations of [cemp] and [cplus].

    Concretely, apply [cplus_emp] wherever possible (symmetrically).
*)
Ltac nocemp :=
  repeat match goal with
    | H: context[@cplus ?l cemp ?x] |- _ => rewrite (@cplus_emp l x) in H
    | |- context[@cplus ?l cemp ?x]      => rewrite (@cplus_emp l x)
    | H: context[@cplus ?l ?X cemp] |- _ => rewrite (@cplus_comm l X cemp) in H
    | |- context[@cplus ?l ?X cemp]      => rewrite (@cplus_comm l X cemp)
  end.

Ltac noemp := nopmemp; nocemp; nonothing.


Lemma mcplus_emp: forall L (m: L.(mach)) c
    (RL: real _ c m),
  mcplus _ m cemp m.
Proof.
  i; econs; noemp; eauto.
Qed.                            



Definition wdet {L} (ma: L.(mach)) := forall evta ca' ma' evtb cf mb mb'
    (STa: step _ evta ma ma')
    (RL: real _ ca' ma')
    (PL: mcplus _ ma cf mb)
    (STb: step _ evtb mb mb'),
  evta = evtb /\ real _ (ca' # cf) mb'. 









Definition anystep {L} m m' :=
  exists evt, step L evt m m'.
Hint Unfold anystep.

Lemma halted_steps: forall LS m m'
    (HA: halted LS m)
    (ST: rtc anystep m m'),
  m = m'.
Proof. clear.
  i; apply clos_rt_rt1n_iff in ST; induction ST; auto.
  rrdes H; eapply halted_step in HA; vauto.
Qed.

Lemma halted_tausteps: forall LS m m'
    (HA: halted LS m)
    (ST: rtc (step _ etau) m m'),
  m = m'.
Proof.
  i; apply clos_rt_rt1n_iff in ST; induction ST; auto.
  eapply halted_step in HA; vauto.
Qed.
