(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import List Arith.
Require Import vgtac basis map.
Require Import lang_common lang_tgt.
Require Import paco2 adequacy.


Definition cheap := option (list (word * word)).
Definition cstack := @tmap word _ (option word).

Structure computer :=
  { cppc: word
  ; cprf: regfile
  ; cpst: cstack
  ; cphp: cheap
  }.

Definition crash (c: computer): computer :=
  {| cppc := cppc c; cprf := cprf c; cphp := None; cpst := cpst c |}.

Fixpoint cloadh (m: module) laddr imports : list (word * word) * list (label * word) :=
  match m with
    | nil => (nil, nil)
    | g :: m' =>
      let ni := length imports in
      let ne := length (g_cptrs g) in
      let nd := length (g_data g) in

      let cptrs := map (fun p => laddr + ni + ne + ne + snd p) (g_cptrs g) in
      let exports := swap (enumerate (laddr + ni + ne) (map fst (g_cptrs g))) in

      let gh := enumerate (laddr + ni + ne) cptrs in
      let lha := enumerate laddr (imports ++ map snd exports) in
      let lhb := enumerate (laddr + ni + ne + ne) (g_data g) in

      let '(h', exports') := cloadh m' (laddr + ni + ne + ne + nd) (imports ++ map snd exports) in
      (lha ++ gh ++ lhb ++ h', exports ++ exports')
  end.

Definition cload (m: module) : option computer :=
  let (h, exports) := cloadh m 0 nil in
  match lookup MAIN exports with
    | Some w =>
      Some {| cppc := match lookup w h with Some a => a | _ => 42 end;
              cprf := [w//r_clo]>>(fun _ => 0);
              cpst := (fun _ => Some 0);
              cphp := Some h
           |}
    | _ => None
  end.


Inductive represents (c: computer) (m: machine) : Prop :=
| rep
    (PC: cppc c = mpc m)
    (RF: cprf c = mrf m)
    (ST: forall a w, cpst c a = Some w <-> pmread (mst m) a w)
    (ST': pmvalid (mst m))
    (HP: forall a w, (exists h, cphp c = Some h /\ lookup a h = Some w) <-> pmread (mhp m) a w)
    (HP': <<VLD: exists l, cphp c = Some l>> <-> <<VLD: pmvalid (mhp m)>>)
.

Lemma represents_finite: forall c m pc rf st hp
    (REP: represents c m)
    (E: m = {| mpc := pc; mrf := rf; mst := st; mhp := hp |})
    (VLD: pmvalid hp),
  pmfinite hp.
Proof.
  i; subst; destruct c; invc REP; ss; clarify.
  apply HP' in VLD; des; clarify.
  exists (map fst l); i.
  hdes; apply HP1 in RD; des; clarify.
  apply in_map_fst; econs.
  eapply list_assoc_someD; eauto.
Qed.

Definition myalloc (rh: raw_heap) (_: word) : word :=
  S (list_max (map fst rh)).

(** This allocator satisfies the axiom from lang_tgt.v. *)
Lemma myalloc_allocator: forall rh i n
    (LT: i < n),
  lookup (i + myalloc rh n) rh = None.
Proof.
  unfold myalloc.
  i; destruct (lookup _ _) eqn:E; vauto.
  exfalso; apply list_assoc_someD in E; apply (in_map fst) in E; ss.
  apply list_max_in in E; omega.
Qed.

Axiom myalloc_is_raw_alloc: raw_alloc = myalloc.

Definition calloc (c: computer) (n: word) : word :=
  match cphp c with
    | Some h => myalloc h n
    | _ => 42                   
  end.

Lemma myalloc_ext: forall rh1 rh2 n
    (EQ: forall l, In l (map fst rh1) <-> In l (map fst rh2)),
  myalloc rh1 n = myalloc rh2 n.
Proof.
  i; unfold myalloc; f_equal; auto using list_max_ext.
Qed.

Lemma calloc_alloc: forall k c m pc rf st hp
    (REP: represents c m)
    (E: m = {| mpc := pc; mrf := rf; mst := st; mhp := hp |}),
  calloc c k = alloc hp k.
Proof.
  i; destruct c, m; inv REP; ss; des; subst.
  unfold alloc, calloc; s; des_ifs.
  - rewrite myalloc_is_raw_alloc; apply myalloc_ext; split; i.
    + destruct (cid _ (pmvalid_some p0)); s.
      destruct (cid _ p); s; subst.
      eapply (list_assoc_map eqdec) in H; des.
      assert (X := HP l0 y); des; hexploit X; [by eauto|]; i.
      dup X1; apply pmread_some in X2.
      hexploit X0; eauto; []; i; des; clarify.
      apply i in X1; apply in_split in X1; des; subst.
      rewrite map_app, flatten_app, map_app.
      apply in_or_app; right; s.
      destruct (x l0) eqn:E; vauto.
    + destruct (cid _ (pmvalid_some p0)); ss.
      destruct (cid _ p); ss; subst.
      apply in_map_iff in H; des; destruct x1; subst; ss.
      apply flatten_inD in H0; des.
      apply in_map_iff in INb; des; subst.
      des_ifs; ss; des; vauto.
      assert (X := HP n n0); des.
      hexploit X0; [by apply pmread_some|]; i; des; clarify.
      apply list_assoc_someD in X2; apply in_map_fst; eauto.
  - exfalso; apply f; eauto.
  - exfalso; apply f; eauto using represents_finite.
  - exfalso; apply VLD0 in p0; des; vauto.
Qed.



Fixpoint extend_heap a n h :=
  match n with
    | O => h
    | S n' => (a, 0) :: extend_heap (S a) n' h
  end.


Definition allocate (c: computer) (n: word) : word * computer :=
  let k := calloc c n in
  (k, {| cppc := cppc c; cprf := cprf c; cpst := cpst c;
         cphp := option_map (extend_heap k n) (cphp c) |}).

Lemma allocate_correct: forall k c m pc rf st hp
    (REP: represents c m)
    (E: m = {| mpc := pc; mrf := rf; mst := st; mhp := hp |})
    (VLD: pmvalid (mhp m)),
  <<LOC: fst (allocate c k) = alloc hp k>> /\
  <<REP': represents (snd (allocate c k))
                     {| mpc := pc; mrf := rf; mst := st;
                        mhp := {alloc hp k |=> repeat 0 k} ## hp |}>>.
Proof.
  unfold allocate; split; i; subst; ss.
  - erewrite calloc_alloc; eauto; ss.
  - inv REP; hdes; ss; subst; dup VLD; apply VLD1 in VLD.
    des; destruct c; ss; clarify.
    assert (VLD: pmvalid ({alloc hp k |=> repeat 0 k} ## hp)).
    { eapply alloc_valid; eauto using represents_finite. }
    econs; auto.
    + erewrite calloc_alloc; eauto.
      generalize dependent (alloc hp k).
      generalize dependent k; induction k; i.
      * split; s; intro IN; des; clarify.
        { ss; hexploit HP0; eauto; []; i.
          eapply pmread_plus; eauto using pmplus_comm.
        }
        { esplits; eauto.
          rewrite pmcells_nil in IN; nopmemp.
          apply HP1 in IN; des; clarify; eauto.
        }
      * s; split; i; des.
        { clarify; ss; rewrite pmcells_cons, <-pmplus_assoc in *.
          des_ifs; [by eapply pmread_plus; eauto|].
          hexploit (IHk (S n)); eauto. 
          { apply pmvalid_plusD in VLD; des; eauto. }
          i; des; hexploit H; eauto; []; i.
          i; eapply pmread_plus; eauto using pmplus_comm.
        }
        { esplits; eauto.
          dup H; apply pmread_valid in H.
          rewrite pmcells_cons, <-pmplus_assoc in H0.
          apply pmread_plusD in H0; des.
          - apply pmread_cellD in H0; des; subst; des_ifs.
          - ss; des_ifs; vauto.
            { exfalso; rewrite pmcells_cons, <-pmplus_assoc in VLD.
              ii; subst; eapply (pmvalidD_plus_read VLD); eauto.
            }
            hexploit IHk; [by eauto using pmread_valid|].
            i; des. eapply H2 in H0; des; clarify.
        }
    + s; esplits; eauto; split; i; des; clarify; eauto.
Qed.


Definition read_pc := cppc.

Definition read_reg := cprf.

Definition read_heap (c: computer) (a: word) : option word :=
  match cphp c with
    | Some h => lookup a h
    | _ => None
  end.

Definition read_stack :=
  cpst.


Definition write_pc (c: computer) (w: word) : computer :=
  {| cppc := w; cprf := cprf c;
     cpst := cpst c; cphp := cphp c |}.

Definition write_reg (c: computer) (r: reg) (w: word) : computer :=
  {| cppc := cppc c; cprf := [w//r]>>(cprf c);
     cpst := cpst c; cphp := cphp c |}.

Definition write_heap (c: computer) (a w: word) : option computer :=
  match cphp c with
    | Some h =>
      match lookup a h with
        | Some _ => Some {| cppc := cppc c; cprf := cprf c;
                            cpst := cpst c; cphp := Some ((a, w)::h) |}
        | _ => None
      end
    | _ => Some c
  end.

Definition write_stack (c: computer) (a w: word) : option computer :=
  match cpst c a with
    | Some _ => Some {| cppc := cppc c; cprf := cprf c;
                        cpst := [Some w//a]>>(cpst c); cphp := cphp c |}
    | _ => None
  end.

Definition write (c: computer) (o: operand) (w: word) : option computer :=
  match o with
    | o_imm _ =>
      None
    | o_reg r =>
      Some (write_reg c r w)
    | o_sind r soff =>
      let addr := add_offset (read_reg c r) soff in
      write_stack c addr w
    | o_hind r soff =>
      let addr := add_offset (read_reg c r) soff in
      write_heap c addr w
  end.


Definition incr_pc c :=
  write_pc c (1 + read_pc c).

Definition evaluate (c: computer) (o: operand) : option word :=
  match o with
    | o_imm w =>
      Some w
    | o_reg r =>
      Some (read_reg c r)
    | o_sind r soff =>
      let addr := add_offset (read_reg c r) soff in
      read_stack c addr
    | o_hind r soff =>
      let addr := add_offset (read_reg c r) soff in
      read_heap c addr
  end.


Inductive eventful_computer :=
| efc_halted
| efc_tau (c: computer)
| efc_in (f: nat -> computer)
| efc_out (n: nat) (c: computer)
.


Definition step' (c: computer) (i: instr) : eventful_computer :=
  match i with
    | i_jmp addr =>
      match evaluate c addr with
        | Some w => efc_tau (write_pc c w)
        | None => efc_tau (crash c)
      end
    | i_jnz r addr =>
      if eqdec (read_reg c r) 0 then efc_tau (incr_pc c) else
        match evaluate c addr with
          | Some w => efc_tau (write_pc c w)
          | None => efc_tau (crash c)
        end
    | i_ld r src =>
      match evaluate c src with
        | Some w => efc_tau (incr_pc (write_reg c r w))
        | None => efc_tau (crash c)
      end
    | i_sto tgt r =>
      match write c tgt (read_reg c r) with
        | Some c' => efc_tau (incr_pc c')
        | None => efc_tau (crash c)
      end
    | i_lpc r =>
      efc_tau (incr_pc (write_reg c r (read_pc c)))
    | i_opr op r arga argb =>
      match evaluate c arga, evaluate c argb with
        | Some wa, Some wb =>
          efc_tau (incr_pc (write_reg c r (bosem op wa wb)))
        | _, _ => efc_tau (crash c)
      end
    | i_alc r rsize =>
      let '(w, c') := allocate c (read_reg c rsize) in
      efc_tau (incr_pc (write_reg c' r w))
    | i_inp r =>
      efc_in (fun n => incr_pc (write_reg c r n))
    | i_out r =>
      efc_out (read_reg c r) (incr_pc c)
  end.


Definition step (c: computer) : eventful_computer :=
  match read_pc c with
    | O => efc_halted
    | _ =>
      match read_heap c (read_pc c) with
        | None => efc_tau (crash c)
        | Some w =>
          match decode_instr w with
            | Some i => step' c i
            | None => efc_tau (crash c)
          end
      end
  end.


Definition comp_step_etau c c' := step c = efc_tau c'.

Inductive cbehmatch behave : computer -> observation -> Prop :=
| cbeh_err c obs
  : cbehmatch behave (crash c) obs
| cbeh_done c
    (DONE : read_pc c = 0)
  : cbehmatch behave c obs_done
| cbeh_inftau c c'
    (ST : step c = efc_tau c')
    (INF: behave c' obs_inftau : Prop)
  : cbehmatch behave c obs_inftau
| cbeh_evt_in c f i obs
    (ST : step c = efc_in f)
    (INF: behave (f i) obs : Prop)
  : cbehmatch behave c (obs_evt (oein i) obs)
| cbeh_evt_out i c c' obs
    (ST : step c = efc_out i c')
    (INF: behave c' obs : Prop)
  : cbehmatch behave c (obs_evt (oeout i) obs)
.

Hint Constructors behmatch.

Inductive cbehave_ cbehave c obs : Prop :=
cbehave_intro c'
  (MAT: cbehmatch cbehave c' obs)
  (TAU : rtc comp_step_etau c c')
: cbehave_ cbehave c obs.
Hint Constructors behave_.

Definition cbehave : _ -> _ -> _ -> Prop := paco2 (@cbehave_).


(** * Correctness. *)

Lemma crashed_correct: forall c m
    (REP: represents c m),
  represents (crash c) (mkerr m).
Proof.
  i; destruct c, m; invc REP; ss; s; subst; econs; vauto.
  s; i; split; i; des; vauto; rrdes H VLD VLD1; vauto.
  split; i; des; ss; rrdes H; vauto.
Qed.

Lemma read_pc_correct: forall c m
    (REP: represents c m),
  read_pc c = mpc m.
Proof.
  i; destruct c, m; invc REP; ss; subst; eauto.
Qed.

Lemma write_pc_correct: forall c m pc rf st hp pc'
    (REP: represents c m)
    (E: m = {| mpc := pc; mrf := rf; mst := st; mhp := hp |}),
  represents (write_pc c pc') {| mpc := pc'; mrf := rf; mst := st; mhp := hp |}.
Proof.
  i; subst; destruct c; unfold write_pc; s; invc REP; eauto.
Qed.

Lemma incr_pc_correct: forall c m pc rf st hp
    (REP: represents c m)
    (E: m = {| mpc := pc; mrf := rf; mst := st; mhp := hp |}),
  represents (incr_pc c) {| mpc := S pc; mrf := rf; mst := st; mhp := hp |}.
Proof.
  i; unfold incr_pc; erewrite read_pc_correct; eauto.
  subst; eauto using write_pc_correct.
Qed.

Lemma write_reg_correct: forall c m pc rf st hp r w
    (REP: represents c m)
    (E: m = {| mpc := pc; mrf := rf; mst := st; mhp := hp |}),
  represents (write_reg c r w)
             {| mpc := pc; mrf := [w//r]>>rf; mst := st; mhp := hp |}.
Proof.
  i; subst; destruct c; unfold write_reg; s; invc REP.
  econs; ss; subst; eauto.
Qed.

Lemma read_heap_correct: forall a w h c m
    (REP: represents c m)
    (HP: mhp m = h),
  pmread h a w <-> read_heap c a = Some w.
Proof.
  i; destruct c, m; invc REP; ss; subst; split; i.
  - unfold read_heap; s.
    eapply HP0 in H; des; subst; eauto.
  - unfold read_heap in H; ss; des_ifs.
    specialize (HP0 a w); des; hexploit HP0; eauto.
Qed.

Lemma write_stack_correct: forall a v w c m c' pc rf st hp
    (REP: represents c m)
    (E: m = {| mpc := pc; mrf := rf; mst := {a |-> v} ## st; mhp := hp |})
    (WR: write_stack c a w = Some c'),
  represents c' {| mpc := pc; mrf := rf;
                   mst := {a |-> w} ## st; mhp := hp |}.
Proof.
  i; destruct c, m; inv REP; ss; subst.
  unfold write_stack in WR; des_ifs; ss; econs; s; eauto using pmvalid_plus_cell.
  split; i.
  - destruct (LEM (a = a0)).  
    + subst; tmext_simplH H; clarify.
      eapply pmread_plus; eauto using pmvalid_plus_cell.
    + rewrite tmext_neq in H; eauto.
      apply ST in H; eapply pmread_plusD in H; des.
      { apply pmread_cellD in H; des; vauto. }
      eapply pmread_plus; eauto using pmvalid_plus_cell.
      rewrite pmplus_comm; eauto.
  - dup H; dup H; apply pmread_valid in H0.
    apply pmread_plusD in H; des.
    + apply pmread_cellD in H; des; subst; tmext_simpl; auto.
    + assert (a <> a0).
      { ii; subst; eapply pmvalidD_plus_read in H0; eauto. }
      rewrite tmext_neq; eauto.
      eapply ST; eapply pmread_plus; [exact H| |]; eauto.
      rewrite pmplus_comm; eauto.
Qed.

Lemma write_heap_correct: forall a v w c m c' pc rf st hp
    (REP: represents c m)
    (E: m = {| mpc := pc; mrf := rf; mst := st; mhp := {a |-> v} ## hp |})
    (WR: write_heap c a w = Some c'),
  represents c' {| mpc := pc; mrf := rf;
                   mst := st; mhp := {a |-> w} ## hp |}.
Proof.
  i; destruct c, m; inv REP; ss; subst.
  unfold write_heap in WR; ss; destruct cphp0; ss.
  - des_ifs; ss; econs; s; eauto using pmvalid_plus_cell.
    { split; i; des; clarify.
      + destruct (LEM (a = a0)).  
        * subst; tmext_simplH H; clarify.
          rewrite list_assoc_cons_eq in H0; clarify.
          eapply pmread_plus; eauto using pmvalid_plus_cell.
        * rewrite list_assoc_cons_neq in H0; eauto.
          hdes; hexploit HP0; eauto; []; i.
          eapply pmread_plusD in X; des.
          { apply pmread_cellD in X; des; vauto. }
          eapply pmread_plus; eauto using pmvalid_plus_cell.
          rewrite pmplus_comm; eauto.
      + dup H; dup H; apply pmread_valid in H0; esplits; eauto.
        apply pmread_plusD in H; des.
        * apply pmread_cellD in H; des; subst; auto.
        * assert (a <> a0).
          { ii; subst; eapply pmvalidD_plus_read in H0; eauto. }
          rewrite list_assoc_cons_neq; eauto.
          hdes; hexploit HP1.
          { eapply pmread_plus; [exact H| |]; eauto.
            rewrite pmplus_comm; eauto.
          }
          i; des; clarify; eauto.
    }
    { split; i; des; clarify; eauto.
      hexploit VLD; eauto using pmvalid_plus_cell.
    }
  - clarify.
    assert (NV: ~ pmvalid ({a |-> v} ## hp)).
    { i; hexploit (HP a v); i; des.
      hexploit X0; eauto using pmread_plus; []; i; des; vauto.
    }
    econs; eauto; s; split; i; des; vauto; exfalso
    ; apply NV; eapply pmvalid_plus_cell; eauto using pmread_valid.
Qed.

Lemma read_reg_correct: forall r c m
    (REP: represents c m),
  read_reg c r = mrf m r.
Proof.
  i; destruct c, m; invc REP; ss; subst; eauto.
Qed.

Lemma evaluate_complete: forall c m pc rf st hp o w
    (REP: represents c m)
    (E: m = {| mpc := pc; mrf := rf; mst := st; mhp := hp |})
    (EV: eval rf st hp o w),
   evaluate c o = Some w.
Proof.
  i; destruct c, m; invc REP; ss; subst.
  unfold evaluate; destruct o; invc EV; s.
  - unfold read_stack; s.
    eapply ST in RD; eauto; des; subst; eauto.
  - unfold read_heap; s.
    eapply HP in RD; eauto; des; subst; eauto.
Qed.

Lemma evaluate_sound: forall c m pc rf st hp o w
    (REP: represents c m)
    (E: m = {| mpc := pc; mrf := rf; mst := st; mhp := hp |})
    (EV: evaluate c o = Some w),
 eval rf st hp o w.
Proof.
  i; destruct c, m; invc REP; ss; subst.
  unfold evaluate in EV; destruct o; vauto; ss.
  - unfold read_stack in EV; ss; des_ifs.
    hexploit (ST (add_offset (rf r) soff) w); i; des.
    hexploit X; eauto.
  - unfold read_heap in EV; ss; des_ifs.
    hexploit (HP (add_offset (rf r) soff) w); i; des.
    hexploit X; eauto.
Qed.

Lemma write_complete: forall c m pc rf st hp dst w m'
    (REP: represents c m)
    (Em: m = {| mpc := pc; mrf := rf; mst := st; mhp := hp |})
    (STO: store pc rf st hp dst w m'),
  exists c',
    <<Ec': Some c' = write c dst w>> /\
    <<REP': represents (incr_pc c') m'>>.
Proof.
  i; destruct c, m; inv REP; ss; subst; invc STO.
  - s; esplits; eauto.
    eapply incr_pc_correct; eauto; eauto using write_reg_correct.
  - s; unfold write_stack; s.
    des_ifs.
    + specialize (ST (add_offset (rf r) soff) n); des.
      hexploit ST; eauto; []; i.
      esplits; eauto.
      eapply incr_pc_correct; eauto.
      eapply write_stack_correct; s; eauto
      ; last by (unfold write_stack; s; rewrite Heq; eauto).
    + exfalso.
      destruct (LEM (pmvalid ({add_offset (rf r) soff |-> v} ## st0))); [|by vauto].
      specialize (ST (add_offset (rf r) soff) v); des.
      hexploit ST0; [by eauto using pmread_plus|]; i; des; vauto.
  - s; unfold write_heap; s.
    destruct cphp0.
    + des_ifs.
      * specialize (HP (add_offset (rf r) soff) n); des.
        hexploit HP; eauto; []; i.
        esplits; eauto.
        eapply incr_pc_correct; eauto.
        eapply write_heap_correct; s; eauto
        ; last by (unfold write_heap; s; rewrite Heq; eauto).
      * exfalso.
        hexploit HP'; eauto; []; i.
        hexploit (HP (add_offset (rf r) soff) v); []; i; des.
        hexploit X1; [by eauto using pmread_plus|]; i; des; vauto.
    + esplits; eauto.
      eapply incr_pc_correct; eauto; inst; f_equal; des.
      assert (NVv: ~ pmvalid ({add_offset (rf r) soff |-> v} ## h)).
      { i; apply VLD0 in FH; des; vauto. }
      assert (NVw: ~ pmvalid ({add_offset (rf r) soff |-> w} ## h)).
      { eauto using pmvalid_plus_cell. }
      destruct ({_|->v} ## _); [by exfalso; unfold pmvalid in *; eapply NVv; vauto|].
      destruct ({_|->w} ## _); [by exfalso; unfold pmvalid in *; eapply NVw; vauto|].
      auto.
Qed.


Lemma store_invalid_heap: forall pc rf st r soff w,
   store pc rf st None (o_hind r soff) w
         {| mpc := S pc; mrf := rf; mst := st; mhp := None |}.
Proof.
  i.
  assert (None = {add_offset (rf r) soff |-> w} ##
                 {add_offset (rf r) soff |-> w}).
  { unfold pmplus; s; des_ifs.
    specialize (o (add_offset (rf r) soff)); des; des_ifs; vauto.
  }
  rewrite H; eauto.
Qed.                            

Lemma write_sound: forall c m pc rf st hp dst w c'
    (REP: represents c m)
    (Em: m = {| mpc := pc; mrf := rf; mst := st; mhp := hp |})
    (Ec': Some c' = write c dst w),
  exists m',
    <<ST: store pc rf st hp dst w m'>> /\
    <<REP: represents (incr_pc c') m'>>.
Proof.
  i; destruct c, m; inv REP; ss; subst.
  destruct dst; vauto; ss.
  - clarify; esplits; eauto.
    eapply incr_pc_correct; eauto; eauto using write_reg_correct.
  - unfold write_stack in Ec'; ss; des_ifs.
    dup Heq; apply ST in Heq; rrdes Heq; subst.
    esplits; eauto.
    eapply incr_pc_correct; eauto.
    eapply write_stack_correct; eauto.
    unfold write_stack; s; rewrite Heq0; auto.
  - unfold write_heap in Ec'; ss; destruct cphp0.
    + hdes; des_ifs; hexploit (HP0 (add_offset (rf r) soff) n); eauto; []; i.
      rrdes X; subst.
      esplits; eauto.
      * econs; s; auto.
        { split; i; des; clarify.
          - apply list_assoc_cons_or in H0; des; subst.
            + eapply pmread_plus; eauto using pmvalid_plus_cell.
            + eapply pmread_plus_cell; eauto.
          - esplits; eauto.
            dup H; apply pmread_plusD in H; des.
            + apply pmread_cellD in H; des; subst; eauto.
            + assert (a <> add_offset (rf r) soff).
              { ii; subst; eapply (pmvalidD_plus_read VLD); eauto. }
              hexploit (HP1 a w0); [by eauto using pmread_plus_cell|].
              i; des; clarify.
              rewrite list_assoc_cons_neq; [|i; omega]; eauto.
        }
        { split; i; des; clarify; eauto using pmvalid_plus_cell. }
    + clarify.
      assert (hp = None).
      { destruct hp; vauto.
        exfalso; hdes; hexploit VLD1; i; des; vauto.
      }
      subst; esplits; [|by eapply incr_pc_correct; eauto].
      eauto using store_invalid_heap.
Qed.


(** ** Completeness wrt. relational language semantics. *)

Lemma step_complete_tau_crash: forall c m
    (REP: represents c m)
    (NZ: mpc m <> 0)
    (ST: forall evt m', ~ step_ok evt m m'),
  step c = efc_tau (crash c).
Proof.
  i; unfold step.
  hexploit read_pc_correct; eauto; []; i; 
  destruct (read_pc c) eqn:P; [by rewrite P in X; omega|].
  destruct (read_heap _ _) eqn:E; [|by auto].
  eapply read_heap_correct in E; eauto.
  hexploit read_pc_correct; eauto; []; i; rewrite X in P.
  rewrite <-P in E; clear X P.
  destruct m; ss; destruct (decode_instr n0) eqn:D; [|by auto].
  apply decode_instr_some in D; rewrite D in E; destruct i; s.
  - destruct (evaluate _ _) eqn:V; [|by auto].
    eapply evaluate_sound in V; eauto.
    exfalso; eapply ST, Sjmp; eauto.
  - des_ifs; exfalso.
    + eapply ST, Sjz; eauto.
      erewrite read_reg_correct in e; eauto; ss.
    + eapply evaluate_sound in Heq; eauto.
      erewrite read_reg_correct in n0; eauto.
      destruct (mrf r) eqn:R; vauto; eapply ST, Sjnz; eauto.
  - destruct (evaluate _ _) eqn:V; [|by auto].
    eapply evaluate_sound in V; eauto.
    exfalso; eapply ST, Sld; eauto.
  - subst; des_ifs.
    erewrite read_reg_correct in Heq; eauto.
    eapply eq_sym, write_sound in Heq; des; eauto.
    exfalso; eapply ST, Ssto; eauto.
  - exfalso; eapply ST, Slpc; eauto.
  - destruct (evaluate _ arga) eqn:Va; [|by auto].
    destruct (evaluate _ argb) eqn:Vb; [|by auto].
    eapply evaluate_sound in Va; eauto.
    eapply evaluate_sound in Vb; eauto.
    exfalso; eapply ST, Sopr; eauto.
  - exfalso; eapply ST, Sinp; eauto.
  - exfalso; eapply ST, Sout; eauto.
  - exfalso; eapply ST, Salc; eauto.
Grab Existential Variables.
  exact 0.
Qed.

Theorem step_complete_tau: forall c m m'
    (ST: tgt_step etau m m')
    (REP: represents c m),
  exists c',
    <<ST': step c = efc_tau c'>> /\ <<REP': represents c' m'>>.
Proof.
  i; invc ST.
  - invc ST0
    ; (hexploit read_pc_correct; eauto; []; i
       ; eapply read_heap_correct in INS; eauto
       ; ss; subst; unfold step, step'; rewrite INS, decode_instr_enum
       ; destruct read_pc; vauto).
    + erewrite evaluate_complete; eauto.
      esplits; eauto using write_pc_correct.
    + hexploit (read_reg_correct r); eauto; []; i.
      ss; rewrite X, EVALV; s.
      esplits; eauto using incr_pc_correct.
    + hexploit (read_reg_correct r); eauto; []; i.
      ss; rewrite X, EVALV; s.
      erewrite evaluate_complete; eauto.
      esplits; eauto using write_pc_correct.
    + erewrite evaluate_complete; eauto; esplits; eauto.
      eapply incr_pc_correct; eauto; eauto using write_reg_correct.
    + eapply write_complete in UPD; eauto; []; i; des.
      erewrite read_reg_correct; eauto; s; auto.
      rewrite <-Ec'; esplits; eauto.
    + esplits; eauto.
      eapply incr_pc_correct; eauto; eauto using write_reg_correct.
    + eapply evaluate_complete in EVALa; eauto; []; rewrite EVALa.
      eapply evaluate_complete in EVALb; eauto; []; rewrite EVALb.
      esplits; eauto.
      invc UPD; eapply incr_pc_correct; eauto; eauto using write_reg_correct.
    + hexploit (read_reg_correct rsize); eauto; []; i.
      ss; rewrite X; esplits; eauto; inv REP; destruct c; ss; clarify.
      unfold read_heap in INS; des_ifs; ss; clarify.
      hdes; hexploit VLD0; eauto; []; i; s.
      eapply incr_pc_correct; eauto.
      erewrite calloc_alloc; eauto.
      eapply write_reg_correct; eauto.
      hexploit (allocate_correct (rf rsize)); eauto; i; des; ss.
      erewrite calloc_alloc in REP'; eauto.
  - eapply step_complete_tau_crash in ST0; eauto.
    esplits; eauto using crashed_correct.
Qed.

Theorem step_complete_in: forall c m n m'
    (ST: tgt_step (ein n) m m')
    (REP: represents c m),
  exists f,
    <<ST': step c = efc_in f>> /\ <<REP': represents (f n) m'>>.
Proof.
  i; hexploit read_pc_correct; eauto; []; i.
  invc ST ST0; ss; subst.
  exists (fun n => incr_pc (write_reg c r n)).
  eapply read_heap_correct in INS; eauto; split.
  - unfold step; destruct read_pc; vauto.
    unfold step'; rewrite INS, decode_instr_enum; auto.
  - eapply incr_pc_correct; eauto.
    eauto using write_reg_correct.
Qed.

Theorem step_complete_out: forall c m n m'
    (ST: tgt_step (eout n) m m')
    (REP: represents c m),
  exists c',
    <<ST': step c = efc_out n c'>> /\ <<REP': represents c' m'>>.
Proof.
  i; hexploit read_pc_correct; eauto; []; i.
  invc ST ST0; ss; subst.
  hexploit (read_reg_correct r); eauto; []; i.
  eapply read_heap_correct in INS; eauto.
  unfold step, step'; rewrite INS, decode_instr_enum; auto.
  rewrite X; esplits; eauto using incr_pc_correct.
  destruct read_pc; vauto.
Qed.


(** ** Soundness wrt. relational language semantics. *)

Lemma step'_sound_tau: forall c i c' m
    (ST: step' c i = efc_tau c')
    (REP: represents c m)
    (NZ: mpc m <> 0)
    (PC: read_heap c (mpc m) = Some (enc i)),
  exists m',
    <<ST': tgt_step etau m m'>> /\
    <<REP': represents c' m'>>.
Proof.
  i; eapply read_heap_correct in PC; eauto.
  destruct m, i; vauto; ss.
  - destruct (evaluate _ _) eqn:E.
    + invc ST.
      eapply evaluate_sound in E; eauto.
      esplits; eauto using write_pc_correct.
      eapply Sok; [by vauto|]; eapply Sjmp; eauto.
    + invc ST.
      esplits; [|by eauto using crashed_correct].
      eapply Serr; [by vauto|]; i.
      invc FH; (hexploit (pmread_uniq INS PC); i; apply enc_inj in X; clarify).
      eapply evaluate_complete in EVAL; eauto; vauto.
  - des_ifH ST.
    + erewrite read_reg_correct in e; eauto; ss; clarify.
      esplits; [by vauto|by eauto using incr_pc_correct].
    + erewrite read_reg_correct in n; eauto; ss; clarify.
      des_ifs.
      * eapply evaluate_sound in Heq; eauto.
        esplits; [|by eauto using write_pc_correct].
        destruct (mrf r) eqn:R; vauto.
      * esplits; [|by eauto using crashed_correct].
        eapply Serr; [by vauto|]; i.
        invc FH; (hexploit (pmread_uniq INS PC); i; apply enc_inj in X; clarify).
        eapply evaluate_complete in EVALA; eauto; vauto.
  - des_ifs.
    + eapply evaluate_sound in Heq; eauto; esplits.
      * eapply Sok, Sld; eauto.
      * eapply incr_pc_correct; [eapply write_reg_correct|]; eauto.
    + esplits; [|by eauto using crashed_correct].
      eapply Serr; [by vauto|]; i.
      invc FH; (hexploit (pmread_uniq INS PC); i; apply enc_inj in X; clarify).
      eapply evaluate_complete in EVAL; eauto; vauto.
  - invc ST; des_ifs.
    + erewrite read_reg_correct in Heq; eauto; ss; clarify.
      eapply eq_sym, write_sound in Heq; eauto; des.
      esplits; eauto; eapply Sok, Ssto; eauto.
    + erewrite read_reg_correct in Heq; eauto; ss; clarify.
      esplits; [|by eauto using crashed_correct].
      eapply Serr; [by vauto|]; i.
      invc FH; (hexploit (pmread_uniq INS PC); i; apply enc_inj in X; clarify).
      eapply write_complete in UPD; eauto; des.
      rewrite Heq in Ec'; vauto.
  - invc ST; esplits.
    + eapply Sok, Slpc; eauto.
    + eapply incr_pc_correct; eauto.
      erewrite read_pc_correct; eauto; eauto using write_reg_correct; eauto.
  - des_ifs.
    + eapply evaluate_sound in Heq; eauto.
      eapply evaluate_sound in Heq0; eauto.
      esplits.
      * eapply Sok, Sopr; eauto.
      * eapply incr_pc_correct; [eapply write_reg_correct|]; eauto.
    + esplits; [|by eauto using crashed_correct].
      eapply Serr; [by vauto|]; i.
      invc FH; (hexploit (pmread_uniq INS PC); i; apply enc_inj in X; clarify).
      eapply evaluate_complete in EVALb; eauto; vauto.
    + esplits; [|by eauto using crashed_correct].
      eapply Serr; [by vauto|]; i.
      invc FH; (hexploit (pmread_uniq INS PC); i; apply enc_inj in X; clarify).
      eapply evaluate_complete in EVALa; eauto; vauto.
  - erewrite read_reg_correct in ST; eauto; ss; clarify.
    hexploit (allocate_correct (mrf rsize)); eauto; []; i; des.
    des_ifs; ss; subst.
    esplits.
    + eapply Sok, Salc; eauto.
    + eapply incr_pc_correct; eauto.
      eapply write_reg_correct; eauto.
      erewrite calloc_alloc; eauto.
Qed.

Theorem step_sound_tau: forall m c c'
    (ST: step c = efc_tau c')
    (REP: represents c m),
  exists m',
    <<ST': tgt_step etau m m'>> /\ <<REP': represents c' m'>>.
Proof.
  i; unfold step in ST.
  hexploit read_pc_correct; eauto; []; i.
  destruct m; rewrite X in ST; clear X; ss.
  destruct mpc; [by vauto|].
  destruct (read_heap _ _) eqn:R.
  - destruct (decode_instr _) eqn:D.
    + apply decode_instr_some in D; subst.
      eapply step'_sound_tau; eauto; s; omega.
    + esplits; [|by invc ST; eauto using crashed_correct].
      eapply Serr; [s; omega|]; i; invc FH
      ; eapply read_heap_correct in INS; eauto; inst; rewrite R in INS
      ; vauto; rewrite decode_instr_enum in D; vauto.
  - esplits; [|by invc ST; eauto using crashed_correct].
    eapply Serr; [s; omega|]; i; invc FH
    ; eapply read_heap_correct in INS; eauto; inst; rewrite R in INS; vauto.
Qed.

Theorem step_sound_in: forall m c f
    (ST: step c = efc_in f)
    (REP: represents c m),
  forall n, exists m',
    <<ST': tgt_step (ein n) m m'>> /\ <<REP': represents (f n) m'>>.
Proof.
  i; hexploit read_pc_correct; eauto; []; i.
  unfold step in ST; rewrite X in ST; clear X.
  destruct (read_heap _ _) eqn:R; destruct (mpc _) eqn:P; vauto.
  eapply read_heap_correct in R; eauto.
  destruct (decode_instr _) eqn:D; [|by vauto].
  apply decode_instr_some in D; subst.
  unfold step' in ST; destruct i; try destruct (evaluate _ _); des_ifs; vauto.
  destruct m; subst; esplits; eauto.
  - econs; s; eauto; [i; ss; omega|].
    s; eapply Sinp; s in P; subst; eauto.
  - eapply incr_pc_correct; eauto; eauto using write_reg_correct.
Qed.

Theorem step_sound_out: forall m c n c'
    (ST: step c = efc_out n c')
    (REP: represents c m),
  exists m',
    <<ST': tgt_step (eout n) m m'>> /\ <<REP': represents c' m'>>.
Proof.
  i; hexploit read_pc_correct; eauto; []; i.
  unfold step in ST; rewrite X in ST; clear X.
  destruct (read_heap _ _) eqn:R; destruct (mpc _) eqn:P; vauto.
  eapply read_heap_correct in R; eauto.
  destruct (decode_instr _) eqn:D; [|by vauto].
  apply decode_instr_some in D; subst.
  unfold step' in ST; destruct i; try destruct (evaluate _ _); des_ifs; vauto.
  destruct m; ss; subst; esplits; eauto.
  - econs; s; eauto; [i; omega|].
    hexploit read_reg_correct; eauto; []; inst; s; i.
    rewrite H; ss; eapply Sout; eauto.
  - eapply incr_pc_correct; eauto; eauto using write_reg_correct.
Qed.


(** ** Soundness of loading modules *)


Lemma pmread_cell_plus {X} {M : EqDec X} {Y : Type} (x : X) (y y' : Y) h : 
  pmvalid ({x |-> y} ## h) -> pmread ({x |-> y} ## h) x y' -> y' = y.   
Proof.
  intros.
  eapply pmread_plusD in H0; des.
  - eapply pmread_cellD in H0; des; auto.
  - exfalso.
    eapply pmvalidD_plus_read; eauto.
Qed.
    
Lemma pmread_plus2 {X} {M : EqDec X} {Y : Type} (x : X) (y : Y) h1 h2 : 
  pmvalid (h1 ## h2) ->
  pmread h2 x y -> pmread (h1 ## h2) x y.
Proof.
  intros. econs; auto. rrdes H0; eauto. eexists.
  rewrite pmplus_comm, RD, pmplus_assoc. reflexivity.
Qed.

Lemma pmread_cells_enumerate {X Y: Type} {M : EqDec X} start l a (w : Y) :
  pmread { start |=> l } a w <-> lookup a (enumerate start l) = Some w.
Proof.
  revert start; induction l.
  - split; i; [ | now inversion H ]. 
    eapply pmcells_readD in H; des.
    now eapply nth_error_someD in NTH.
  - split; i.
    + simpl. destruct eq_nat_dec; subst.
      * eapply pmcells_readD in H; des. 
        rewrite NPeano.Nat.sub_diag in NTH.
        now inversion NTH.
      * eapply IHl; eapply pmread_plusD in H; des; eauto.
        eapply pmread_cellD in H; des; subst; exfalso; now auto.
    + simpl in H. destruct eq_nat_dec in H; subst.
      * eapply (pmcells_read 0); auto.
      * assert (pmvalid { start |=> a0 :: l }) by eapply pmcells_valid. 
        rewrite pmcells_cons in *.
        eapply pmread_plus2; eauto.
        now eapply IHl.
Qed.

Lemma lookup_app {X : Type} {M : EqDec X} {Y : Type} xs {xs'} x (v : Y) :
  lookup x (xs ++ xs') = Some v -> lookup x xs = Some v \/ lookup x xs' = Some v.
Proof.
  induction xs; eauto.
  destruct a; simpl. destruct eqdec; eauto.
Qed.

Lemma app_lookup {X : Type} {M : EqDec X} {Y : Type} xs {xs'} x (v : Y) :
  lookup x xs = Some v \/ (lookup x xs = None) /\ lookup x xs' = Some v -> lookup x (xs ++ xs') = Some v.
Proof.
  induction xs; i; des; try (now vauto); destruct a; simpl in *; destruct eqdec; try now eauto.
Qed.

Lemma pmread_plusD_strong {X} {M : EqDec X} {Y : Type} (x : X) (y : Y) h1 h2 : 
  pmread (h1 ## h2) x y -> pmread h1 x y \/ ((forall y, pmread h1 x y -> False) /\ pmread h2 x y).
Proof.
  intros.
  assert (pmvalid (h1 ## h2)) by eauto.
  eapply pmread_plusD in H; des; eauto.
  right; split; eauto; i.
  eapply pmvalidD_plus_read; now eauto. 
Qed.

Lemma etau_steps_comp m m' c :
  rtc (tgt_step etau) m m' ->
  represents c m -> 
  exists c',
    rtc comp_step_etau c c' /\ represents c' m'.
Proof.
  intros RTC; revert c.
  induction RTC; i; eauto.
  - eapply step_complete_tau in H0; eauto; des; eexists; splits; [| eassumption]; eauto.   
  - eapply IHRTC1 in H; des.
    eapply IHRTC2 in H0; des.
    eexists; splits; [| eassumption]. eapply Relation_Operators.rt_trans; now eauto.
Qed.

Lemma etau_steps_mach c c' m :
  rtc (comp_step_etau) c c' ->
  represents c m -> 
  exists m',
    rtc (tgt_step etau) m m' /\ represents c' m'.
Proof.
  intros RTC; revert m.
  induction RTC; i; eauto.
  - eapply step_sound_tau in H0; eauto; des; eexists; splits; [| eassumption]; eauto.   
  - eapply IHRTC1 in H; des.
    eapply IHRTC2 in H0; des.
    eexists; splits; [| eassumption]. eapply Relation_Operators.rt_trans; now eauto.
Qed.

Definition dummy_conf (m : machine) : config := 
  {| cpc := Just (mpc m)
   ; crf := Just (mrf m) 
   ; cst := mst m
   ; chp := mhp m |}.


Lemma error_inv c m :
  represents c m ->
  lang_spec.error tgt_lsi m -> mhp m = impossible.
Proof.
  ii. specialize (H0 (dummy_conf m)). 
  destruct (mhp m) eqn:EQ; eauto.
  exfalso; eapply H0.
  econs; splits; eauto; try rewrite EQ; try now clarify.
  - now invc H.
  - invc H. 
    eapply (represents_finite c m (mpc m) (mrf m) (mst m)); eauto; clarify.
    destruct m. rewrite <- EQ. eauto. 
Qed.

Lemma error_crash (m : lang_spec.mach tgt_lsi) c :
  represents c m ->
  lang_spec.error tgt_lsi m ->
  exists c',
    c = crash c'.
Proof.
  ii. eapply error_inv in H0; eauto.
  invc H; des.
  exists (crash c).
  destruct c, m. unfold crash. simpl. f_equal.
  destruct cphp0; auto.
  simpl in *. assert (pmvalid mhp) by (eapply VLD; eauto).
  rewrite H0 in H. clarify. exfalso; eauto.
Qed.

Lemma crash_error c m :
  represents (crash c) m ->
  lang_spec.error tgt_lsi m.
Proof.
  ii. simpl in FH. invc FH. invc H; des. 
  eapply VLD0 in HV; des; now vauto.
Qed.


Lemma tgt_wfmod_cons: forall g m oenv
    (WF: tgt_wfmod (g :: m) oenv),
  exists oenv',
    <<WF: tgt_wfmod m oenv'>>.
Proof.
  i; destruct g; unfold tgt_wfmod.

  assert (exists oenv',
            oenv' = flatten (map (map fst <*> lang_tgt.g_cptrs) m)).
  {
    rrdes WF; clear BND; ginduction g_cptrs; vauto.
  }

  des; econs; esplits; auto.
  i; eapply WF; ss; eauto.
Qed.

Lemma cload_wf: forall mod n im h exports l v oenv
    (WF: tgt_wfmod mod oenv)
    (L: cloadh mod n im = (h, exports))
    (EX: lookup l exports = Some v),
  exists w, lookup v h = Some w.
Proof.
  induction mod; i; ss; clarify.
  destruct (cloadh _ _ _) eqn:E; clarify.
  apply list_assoc_app_or in EX; des.
  - rewrite !list_assoc_app; des_ifs; eauto.
    exfalso.
    apply list_assoc_someD, swap_inD, enumerate_inD in EX; des.

    assert (Heq': exists w, nth_error (map
        (fun p => n + length im + length (g_cptrs a) + length (g_cptrs a) + snd p)
        (g_cptrs a)) (v - (n + length im + length (g_cptrs a))) = Some w).
    { eapply nth_error_some'; rewrite map_length in *; omega. }

    des; eapply pmcells_read in Heq'.
    + apply pmread_cells_enumerate in Heq'; ss.
      rewrite Heq' in Heq0; vauto.
    + omega.
  - rewrite !list_assoc_app; des_ifs; eauto.
    apply tgt_wfmod_cons in WF; des.
    eapply IHmod in E; [|by ii; eapply WF0; ss; eauto|]; inst; eauto.
Qed.



Lemma load_wf: forall m laddr imports gh lh exports l v oenv
    (WF: tgt_wfmod m oenv)
    (E: load m laddr imports = (gh, lh, exports))
    (EX: lookup l exports = Some v),
  exists w, pmread gh v w.
Proof.
  i; hexploit load_valid; eauto; []; i.
  ginduction m; i; ss; clarify.
  destruct (load _ _ _) as [[gh' lh']] eqn:E'; clarify.
  apply list_assoc_app_or in EX; des.
  - apply list_assoc_someD, swap_inD, enumerate_inD in EX; des.

    assert (Heq': exists w, nth_error (map
        (fun p => laddr + length imports + length (g_cptrs a) + length (g_cptrs a) + snd p)
        (g_cptrs a)) (v - (laddr + length imports + length (g_cptrs a))) = Some w).
    { eapply nth_error_some'; rewrite map_length in *; omega. }

    des; eapply pmcells_read in Heq'.
    + ss; eauto.
      econs; eapply pmread_plus; eauto.
      decomp_pmvalid X; eauto.
    + omega.
  - decomp_pmvalid X.
    apply tgt_wfmod_cons in WF; des.
    eapply IHm in E'; [|by ii; eapply WF0; s; eauto| |]; inst; eauto.
    + des; econs; eapply pmread_plus; eauto; rewrite pmplus_comm; eauto.
    + rewrite pmplus_comm; eauto.
Qed.



Lemma load_repr_aux: forall mod n im gh lh mex h cex
  (ML: load mod n im = (gh, lh, mex))
  (CL: cloadh mod n im = (h, cex)),
  <<EX: mex = cex>> /\
  <<HP: forall a w, lookup a h = Some w <-> pmread (gh ## lh) a w>>.
Proof.
  i; hexploit load_valid; eauto; []; i.
  ginduction mod; i.
  - ss; clarify; nopmemp; split; auto; split; i; vauto.
    apply pmread_empD in H; vauto.
  - ss; destruct (load _ _ _) as [[gh' lh']] eqn:E.
    destruct (cloadh _ _ _) as [h'] eqn:F.
    clarify.
    rewrite <-!pmplus_assoc; rhs rewrite 2(pmplus_swap gh').
    rewrite <-!pmplus_assoc in X; rhs rewrite 2(pmplus_swap gh') in X.
    
    dup X; decomp_pmvalid X.
    eapply IHmod in F; eauto; clear E IHmod.
    rewrite pmplus_swap in X0 |- *.
    des; clarify; split; auto; split; i.
    + apply list_assoc_app_or in H; des.
      { eapply pmread_plus; first eapply pmread_cells_enumerate; eauto. }
      apply list_assoc_app_or in H0; des.
      { eapply pmread_plus; first eapply pmread_cells_enumerate; eauto.
        (rhs rewrite (pmplus_swap {n |=> im ++ _})); eauto.
      }
      apply list_assoc_app_or in H1; des.
      { eapply pmread_plus; first eapply pmread_cells_enumerate; eauto.
        (rhs rewrite <-2(pmplus_swap {n + _ + _ + _ |=> _})); eauto.
      }
      apply HP in H2.
      eapply pmread_plus; eauto.
      (rhs rewrite <-(pmplus_comm (_ ## lh')), <-!(pmplus_swap (_ ## lh'))); eauto.
    + apply pmread_plusD in H; des.
      { apply pmread_cells_enumerate in H; ss; rewrite list_assoc_app, H; auto. }
      apply pmread_plusD in H; des.
      { dup H; apply pmcells_readD in H0; des.
        apply pmread_cells_enumerate in H; ss; rewrite list_assoc_app.
        des_ifs.
        { exfalso; apply pmread_cells_enumerate, pmcells_readD in Heq; des.
          rewrite swap_snd, app_length, map_length, enumerate_length, map_length in *.
          omega.
        }
        rewrite list_assoc_app, H; auto.
      }
      apply pmread_plusD in H; des.
      { dup H; apply pmcells_readD in H0; des.
        apply pmread_cells_enumerate in H; ss.
        rewrite list_assoc_app; des_ifs.
        { exfalso; apply pmread_cells_enumerate, pmcells_readD in Heq; des.
          rewrite swap_snd, app_length, map_length, enumerate_length, map_length in *.
          omega.
        }
        rewrite list_assoc_app; des_ifs.
        { exfalso; apply pmread_cells_enumerate, pmcells_readD in Heq0; des.
          rewrite !map_length in LE2; omega.
        }
        rewrite list_assoc_app, H; auto.
      }
      rewrite !list_assoc_app; des_ifs.
      * exfalso; apply pmread_cells_enumerate in Heq.
        eapply pmvalidD_plus_read in H; eauto.
        eapply pmvalid_plus_plus; eauto.
      * exfalso; apply pmread_cells_enumerate in Heq0.
        eapply pmvalidD_plus_read in H; eauto.
        eapply pmvalid_plus_plus; eauto.
      * exfalso; apply pmread_cells_enumerate in Heq1.
        eapply pmvalidD_plus_read in H; eauto.
        eapply pmvalid_plus_plus; eauto.
      * apply HP; auto.
Qed.


Lemma load_repr: forall mod m c oenv
    (WF: tgt_wfmod mod oenv)
    (Lm: tgt_load mod = Some m) 
    (Lc: cload mod = Some c),
  represents c m.
Proof.
  ii; assert (VLD := load_valid mod 0 nil).
  unfold tgt_load in *; unfold cload in *.
  destruct (load mod _ _) as [[gh lh]] eqn:E.
  destruct (cloadh mod _ _) as [h] eqn:F.
  hexploit load_repr_aux; eauto; []; i; des; subst.
  destruct (lookup _ _) eqn:L; [|by vauto].
  clarify; econs; s; vauto.
  - dup WF; eapply cload_wf in WF0; eauto; []; des.
    ss; rewrite WF0.
    eapply load_wf in WF; eauto; []; des.
    dup WF; dup WF; apply pmread_valid, pmvalid_some in WF; des; subst.
    apply pmread_some in WF1; rewrite WF1.
    hexploit (HP n w0); i; des.
    rewrite WF0 in X0; hexploit X0; [|by i; vauto].
    eapply pmread_plus; eauto.
  - simpl; intros; split; intros.
    + by eapply pmread_some. 
    + by eapply pmread_some in H.
  - i; split; i; des; clarify.
    + eapply HP; eauto.
    + apply HP in H; des; eauto.
  - eauto 10.
Qed.

Theorem repr_behave_mc :
  forall (m : lang_spec.mach tgt_lsi) c o
         (REP : represents c m)
         (BH : behave bot2 m o),
    << CBH : cbehave bot2 c o >>.
Proof.
  pcofix CIH; i; pfold; destruct o as [| |[]]; clarify;
  invc BH; invc SIM; eapply etau_steps_comp in TAU; eauto; des; econs; eauto; invc MAT;
     try (now eapply error_crash in ERR; eauto; des; subst; econs);
     [ econs; invc TAU0; now invc TM 
     | eapply step_complete_tau in ST
     | eapply step_complete_in in ST
     | eapply step_complete_out in ST]; 
     eauto; des; econs; eauto; eapply LE in INF; pclearbot; now eauto.
Qed.

Theorem repr_behave_cm : 
  forall (m : lang_spec.mach tgt_lsi) c o
         (REP : represents c m)
         (CBH : cbehave bot2 c o),
    << BH : behave bot2 m o >>.
Proof.
  pcofix CIH; i; pfold; destruct o as [| |[]]; clarify;
  invc CBH; invc SIM; eapply etau_steps_mach in TAU; eauto; des; econs; eauto; invc MAT;
     try (now eapply crash_error in TAU0; now eauto);
     [ eapply beh_done; invc TAU0; unfold read_pc in DONE; now rewrite PC in DONE
     | eapply step_sound_tau in ST
     | eapply step_sound_in in ST
     | eapply step_sound_out in ST]; 
     eauto; des; eapply LE in INF; pclearbot; now eauto.
Qed.

Theorem load_behave: forall (m: lang_spec.mach tgt_lsi) c M o oenv
     (WFMOD: tgt_wfmod M oenv)
     (TLOAD: tgt_load M = Some m) 
     (CLOAD: cload M = Some c),
  behave bot2 m o <-> cbehave bot2 c o.
Proof.
  ii; split; [eapply repr_behave_mc | eapply repr_behave_cm]
  ; eapply load_repr; eauto.
Qed.
