(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program Relations Omega.
Require Import paco2  paco9  vgtac.
Require Import basis myclassical countable lang_common lang_spec.
Require Import  model_common  model_untyped .
Require Export model_properties_common.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



(** Least Fixpoint of a function on grel *)


Inductive gstar {La Lb} (w: world La Lb) f b : w.(state) -> @vrelf La Lb :=
| gstar_base s  va vb
  (R: b s  va vb : Prop)
  : gstar w f b s  va vb
| gstar_step g s  va vb 
    (LE: g <g= gstar w f b) (R: f g s  va vb : Prop)
  : gstar w f b s  va vb
.
Hint Constructors gstar.

Lemma gstar_induction La Lb (w: world La Lb): forall f b g
    (MON: forall g g' (LE: g <g= g'), f g <g= f g')
    (TS: b \g/ f g <g= g),
  gstar w f b <g= g.
Proof.
  i; induction PR; auto.
  apply TS; right.
  eapply MON; eauto.
Qed.

Lemma gstar_mon La Lb (w: world La Lb): forall f b b'
    (MON: forall g g' (LE: g <g= g'), f g <g= f g')
    (TS: b <g= b'),
  gstar w f b <g= gstar w f b'.
Proof.
  i; eapply gstar_induction, PR; eauto.
  i; des; eauto.
Qed.



(** Properties of Greatest Respectful Closure *)


Section Properties_of_Greatest_Respectful_Closure.

Variables La Lb : LangSpec.
Variable w : world La Lb.

Lemma gres_resclo:
  resclo (@gres La Lb w).
Proof.
  econs; eauto using gres_mon.
  ii; destructs PR RES; eapply esim_mon with (r:=clo r); eauto.
Qed.

Lemma resclo_compose: forall clo1 clo2 
    (RES1: resclo clo1) (RES2: resclo clo2),
  @resclo La Lb w (clo1 <*> clo2).
Proof.
  i; destruct RES1, RES2; unfold compose; econs; eauto.
Qed.

Lemma gres_incl_clo: forall clo (RES: @resclo La Lb w clo) r,
  clo r <e= gres r.
Proof.
  eauto.
Qed.

Lemma gres_incl_id: forall r,
  r <e= @gres La Lb w r.
Proof.
  i; eexists (fun x => x); eauto.
Qed.
Hint Resolve gres_incl_id.

Lemma gres_idem: forall (r: @esimtyp La Lb w),
  gres (gres r) <e= gres r.
Proof.
  i; eapply gres_incl_clo with (clo := gres <*> gres)
  ; eauto using resclo_compose, gres_resclo.
Qed.

Lemma resclo_sound: forall clo (RES: @resclo La Lb w clo)
    r (PFIX: r <e= esim_ (clo r)),
  r <e=  paco9  esim_ bote.
Proof.
  set (rclo := fix rclo clo n (r: @esimtyp La Lb w) :=
    match n with
    | 0 => r
    | S n' => rclo clo n' r \e/ clo (rclo clo n' r)
    end).
  i; revert_until PFIX.
  cut (forall n, rclo clo n r <e= esim_ (rclo clo (S n) r)).
  { i; assert (R: exists n, rclo clo n r i kia kib g si s sync  ea eb) 
      by (exists 0; eauto); clear PR.
    revert i kia kib g si s sync  ea eb R; pcofix CIH.
    i; pfold; eapply esim_mon; eauto. }
  induction n; ss; i; [by eapply esim_mon; eauto|].
  des; [by eapply esim_mon; eauto|].
  eapply esim_mon; [eapply RES; [|apply IHn|]|]; inst; s; eauto.
Qed.

End Properties_of_Greatest_Respectful_Closure.



(** Basic Properties *)


Section Basic_Properties.

Variables La Lb : LangSpec.
Variable w: world La Lb.

Lemma clear_bote: forall (p: esimtyp w),
  p \e/ bote = p.
Proof.
  repeat (i; apply functional_extensionality).
  i; apply prop_ext; intuition.
Qed.

Lemma pesim_incl_gres: forall r,
  @gres La Lb w (pesim r) <e= pesim r.
Proof.
  intro r; pcofix CIH; i; pfold.
  eapply esim_mon, gres_idem.
  destruct (gres_resclo w). 
  eapply RESPECT, PR; i; [by eauto using gres_incl_id|].
  punfold PR0.
  eapply esim_gres_mon; eauto. 
  i; destruct PR1; eauto.
  by left; eapply pesim_mon; eauto.
Qed.

Lemma esim_fold:
  esim_ (esim w) <e= esim w.
Proof.
  i; pfold; eapply esim_mon; eauto using gres_incl_id.
Qed.

Lemma esim_unfold:
  esim w <e= esim_ (esim w).
Proof.
  i; exploit (resclo_sound (gres_resclo w) (esim w)); eauto.
  { i; punfold PR0; rewrite clear_bote in PR0; eauto using esim_mon. }

  clear PR; intro PR; punfold PR; rewrite clear_bote in PR.
  eapply esim_mon; eauto.
  clear -La; s.
  pcofix CIH.
  i; pfold; punfold PR; rewrite clear_bote in PR.
  eapply esim_mon; eauto using gres_incl_id.
Qed.

Lemma call_ptwise: forall s rf rv rk  ea eb
    (STK: call w s rf rv rk  ea eb),
  exists  fa fb,
    call w s (vsingle  fa fb) rv rk  ea eb /\
    rf  fa fb.
Proof.
  i; destruct STK; depdes RF; esplits; (s; eauto).
Qed.

Lemma call_union: forall s
  (rf rfa rfb: vrelf) rv rk  ea eb
    (STK: call w s rf rv rk  ea eb)
    (UNION: forall  va vb 
              , 
            rf  va vb -> (rfa \r/ rfb)  va vb),
  call w s rfa rv rk  ea eb \/
  call w s rfb rv rk  ea eb.
Proof.
  i; apply call_ptwise in STK; des.
  apply UNION in STK0 .
  des.
  - left; eapply call_mon; eauto.
    by i; depdes PR.
  - right; eapply call_mon; eauto.
    by i; depdes PR.
Qed.

Lemma pesim_release: forall r i kia kib g si s
     (ea: La.(conf)) (eb: Lb.(conf)) hsa hsb
    (ES: pesim r i kia kib g si s true  ea eb)
    (HR: w.(confrel) g s hsa hsb)
    (COa: core ea)
    (COb: core eb),
  pesim r i kia kib g si s false  (ea # hsa) (eb # hsb).
Proof.
  i; punfold ES; pfold.
  i; destruct ES; econs; i.
  destruct CNF; des; subst.
  rewrite cplus_emp, <-cplus_assoc in MA, MB.
  exploit ESIM; eauto.
Qed.

Lemma pesim_acquire: forall r i kia kib g si s
     (ea: La.(conf)) (eb: Lb.(conf))
    (ES: forall hsa hsb (HR: w.(confrel) g s hsa hsb),
           pesim r i kia kib g si s false  (ea # hsa) (eb # hsb)),
  pesim r i kia kib g si s true  ea eb.
Proof.
  i; pfold; econs; i; inv CNF; exploit ES; eauto.
  i; punfold x0; inv x0; eapply ESIM; eauto.
  rewrite cplus_assoc in MA; rewrite cplus_assoc in MB.
  econs; eauto; noemp; ss; eauto.
Qed.

(** It is safe to assume [real] when proving [pesim]. *)
Lemma pesim_real: forall r i kia kib g si s
     (ea: La.(conf)) (eb: Lb.(conf))
    (SIM: forall ma mb (RLa: real _ ea ma) (RLb: real _ eb mb),
            pesim r i kia kib g si s false  ea eb),
  pesim (w:=w) r i kia kib g si s false  ea eb.
Proof.
  i; pfold; econs; i; inv CNF; des; subst; noemp.
  hexploit (extra_real _ ea); i; des.
  hexploit (extra_real _ eb); i; des.
  exploit SIM; eauto; i.
  punfold x0; inv x0; eauto.
Qed.  

(** It is safe to assume [gknow_] when proving [pesim]. *)
Lemma pesim_gknow_: forall r i kia kib g si s sync
     (ea: La.(conf)) (eb: Lb.(conf))
    (SIM: forall (GK: gknow_ w (gres (pesim r \e/ r)) g),
            pesim r i kia kib g si s sync  ea eb),
  pesim (w:=w) r i kia kib g si s sync  ea eb.
Proof.
  i; pfold; econs; i; exploit SIM; eauto.
  i; punfold x0; eapply x0; eauto.
Qed.

(** It is safe to assume [gknow] when proving [pesim]. *)
Lemma pesim_gknow: forall r i kia kib g si s sync
     (ea: La.(conf)) (eb: Lb.(conf))
    (SIM: forall (GK: gknow w g),
            pesim r i kia kib g si s sync  ea eb),
  pesim (w:=w) r i kia kib g si s sync  ea eb.
Proof.
  i; apply pesim_gknow_; i; apply SIM.
  eapply gknow_ant; eauto; i; eapply gres_incl_id; left.
  eapply  paco9_mon ; eauto; i; vauto.
Qed.


(** [pesim] is closed under silent expansions on the source side. *)

Lemma pesim_stepb: forall (r: esimtyp w)
    i i' kia kib g si s  ea eb
    (ST: forall cfb mb (XTR: extra cfb) (RL: real _ (eb # cfb) mb),
           exists eb' mb', <<ST: tc (step _ etau) mb mb'>> /\
                       (<<ERR: error _ mb'>> \/
                        (<<RL: real _ (eb' # cfb) mb'>> /\
                         <<SIM: pesim r i' kia kib g si s false  ea eb'>>))),
  pesim r i kia kib g si s false  ea eb.
Proof.
  i; pfold; econs; i.
  inv CNF; des; subst; noemp.
  exploit ST; eauto; i; des.
  { econs 1; [ apply tc_rtc|]; eauto. }
  punfold SIM; invc SIM; exploit ESIM; [M|M|M| |]; (Mdo by eauto)
  ; first by econs; eauto; noemp; eauto.
  i; des; inv x0; [M| |M]; Mdo by eauto using rt_trans, tc_rtc, ext_trans.
  eapply esim_step; eauto; i.
  hexploit EST; eauto; []; i; des.
  - inv ES; des; subst.
    + esplits
      ; first (by eapply rt_trans; first (eapply rt_trans; [eapply tc_rtc|]); eauto)
      ; eauto using ext_trans.
    + hexploit tc_last
      ; first by eapply tc_rtc_tc; [exact ST0|by eapply rt_trans; eauto].
      i; des; esplits; eauto using ext_trans.
  - esplits
      ; first (by eapply rt_trans; first (eapply rt_trans; [eapply tc_rtc|]); eauto)
      ; eauto using ext_trans.
Qed.

Lemma pesim_stepb_obsolete: forall (r: esimtyp w)
    i i' kia kib g si s  ea eb eb'
    (ST: forall cfb mb (XTR: extra cfb) (RL: real _ (eb # cfb) mb),
           exists mb', <<ST: tc (step _ etau) mb mb'>> /\
                       <<RL: real _ (eb' # cfb) mb'>>)
    (SIM: pesim r i' kia kib g si s false  ea eb'),
  pesim r i kia kib g si s false  ea eb.
Proof.
  i; pfold; econs; i.
  inv CNF; des; subst; noemp.
  exploit ST; eauto; i; des.
  punfold SIM; invc SIM; exploit ESIM; [M|M|M| |]; (Mdo by eauto)
  ; first by econs; eauto; noemp; eauto.
  i; des; inv x0; [M| |M]
  ; Mdo by eauto using rt_trans, tc_rtc.
  eapply esim_step; eauto; i.
  hexploit EST; eauto; i; des.
  - inv ES; des; subst.
    + esplits
      ; first (by eapply rt_trans; first (eapply rt_trans; [eapply tc_rtc|]); eauto)
      ; eauto.
    + hexploit tc_last
      ; first by eapply tc_rtc_tc; [exact ST0|by eapply rt_trans; eauto].
      i; des; esplits; eauto.
  - esplits
      ; first (by eapply rt_trans; first (eapply rt_trans; [eapply tc_rtc|]); eauto)
      ; eauto.
Qed.


End Basic_Properties.


(** Context Closure *)


(** 
  Definition 
 *)

Inductive ksimid {La Lb} w : @ksimtyp La Lb w :=
| _ksimid: forall g si sp s  ka kb
     (EXT: w.(extpub) sp si), 
   ksimid w wfo_zero ka kb g si sp s  ka kb 
.
Hint Constructors ksimid.

Inductive compkc {La Lb w} (ks : ksimtyp w) (es : esimtyp w) : @esimtyp La Lb w :=
_compkc: forall i j kia kib 
  g si sp sync  ka kb s ea eb
    (RC: es i ka kb g sp s sync  ea eb)
    (RK: ks j kia kib g si sp s  ka kb),
  compkc ks es (wfo_add j i) kia kib g si s sync  ea eb.
Hint Constructors compkc.

Inductive compkk {La Lb w} (kso ksi : ksimtyp w) : @ksimtyp La Lb w :=
_compkk: forall i j kia kib 
  si sm g sp s  kma kmb ka kb
    (RKI: ksi i kma kmb g sm sp s  ka kb)
    (RKO: kso j kia kib g si sm s  kma kmb),
  compkk kso ksi (wfo_add j i) kia kib g si sp s  ka kb.
Hint Constructors compkk.

Inductive kstar {La Lb w} (p: @esimtyp La Lb w) 
  i kia kib g si sp s  ka kb : Prop 
:=
| kstar_id 
  (R: ksimid w i kia kib g si sp s  ka kb)
| kstar_app 
  (R: compkk (kstar p) (ksim_ p) i kia kib g si sp s  ka kb)
.  
Hint Constructors kstar.

Inductive ksimsingle {La Lb} (w: world La Lb) i kia kib
  g si sp s  ka kb : ksimtyp w :=
| _ksimsingle s' (EXT: w.(ext) s' s) g' (GE: g <g= g')
  : ksimsingle w i kia kib g si sp s  ka kb 
                 i kia kib g' si sp s'  ka kb 
.
Hint Constructors ksimsingle.

Definition ctxclo {La Lb w} (p: @esimtyp La Lb w) := compkc (kstar p) p.
Hint Unfold ctxclo.

(**
   Context Closure is respectful 
 *)

Section Context_Closure_Is_Respectful.

Variables La Lb : LangSpec.
Variable w : world La Lb.

Lemma kstar_induction: forall p (r: ksimtyp w)
    (TS: ksimid w \k/ compkk r (ksim_ p) <k= r),
  kstar p <k= r.
Proof.
  intros until i; revert i;  fix 10 .
  i; destruct PR; [by apply TS; left; apply R|].
  destruct R; apply TS; right. 
  econs; i; eauto.
Qed.

Lemma kstar_mon: forall p i kia kib 
  g si sp s  ka kb
    (R: kstar p i kia kib g si sp s  ka kb)
    p'  (LEX: p <e= p')
    g'  (LEG: g <g= g')
    si' (EXTI: w.(extpub) si si')
    sp' (EXTP: w.(extpub) sp' sp)
    s'  (EXT: w.(ext) s' s),
  kstar p' i kia kib g' si' sp' s'  ka kb.
Proof. 
  i; revert g' si' sp' s' LEG EXTI EXTP EXT.
  pattern i, kia, kib, g, si, sp, s,  ka, kb.
  eapply kstar_induction, R; inst; clear -LEX; i; des.
  - depdes PR; econs 1; econs; eauto using extpub_trans, ext_trans.
  - destruct PR; econs 2; econs; eauto using ksim_mon, ext_trans, vquerya_mon, vqueryb_mon.
Qed.

Lemma kstar_kctx: forall (p: esimtyp w),
  compkk (kstar p) (ksim_ p) <k= ksim_ (ctxclo p).
Proof.
  i; destruct PR; destruct RKI.
  econs; eauto using kstar_mon.
Qed.



Lemma ctxclo_incl: forall (p: esimtyp w), p <e= ctxclo p.
Proof.
  i; rewrite <- (w.(wfo_add_zero_l) i); eauto.
Qed.

Lemma esim_main_taus: forall esimstep
  kia kib g si s  cfa ma cfb mb mb'
    (R: @esim_main La Lb w esimstep kia kib 
          g si s  cfa ma cfb mb')
    (TAUS: rtc (Lb.(step) etau) mb mb'),
  esim_main esimstep kia kib g si s  cfa ma cfb mb.
Proof.
  i; destruct R.
  - econs 1; try eapply rt_trans; eauto.
  - econs 2; eauto. 
    i; apply EST in ST; des.
    esplits; try eapply ES; eauto using rt_trans.
  - econs 3; try eapply rt_trans; eauto.
Qed.

Lemma key_lemma: forall (px py: esimtyp w)
    (LEA: px <e= py)
    (LEB: px <e= esim_ py),
  compkc (kstar px) (esim_ py) <e= esim_ (ctxclo py).
Proof.
  i; set (ihyp i kia kib g si sp s  ka kb :=
          <<KS: kstar px i kia kib g si sp s  ka kb>> /\
          <<KC: compkc (ksimsingle w i kia kib g si sp s  ka kb) (esim_ py) 
                <e= esim_ (ctxclo py)>>).

  cut (kstar px <k= ihyp).
  { intro X; destruct PR; apply X in RK; apply RK; eauto. }
  
  clear -LEA LEB; i.
  dup PR.
  eapply kstar_induction, PR; inst; clear -LEA LEB; i.

  assert (KST: kstar px i kia kib g si sp s  ka kb).
  { des; depdes PR; [by eauto using extpub_trans|].
    by econs 2; econs; [|eapply RKO]; eauto using ksim_mon. }

  econs; [by apply KST|].

  guardH PR.
  r; ii; depdes PR0 RK.

  destruct RC; econs; eauto.
  i; exploit ESIM. 
  - eapply gknow_ant; eauto; inst; eauto using ctxclo_incl.
  - eauto using confrel_mon, vquerya_mon, vqueryb_mon, kstar_mon.
  - eauto.
  - eauto.
  - clear ESIM .
    intro MAIN; destruct MAIN.
    + 
      eauto.
    + 
      eapply esim_step; eauto.
      ii; specialize (EST _ _ ST); des; depdes ES.
      * 
        {
        esplits; eauto.
        left; econs; eauto.
        - des; eauto.
          subst; right; esplits; eauto.
          eapply wfo_add_ltc2, LT.
        - econs.
          + ii; eapply ES; i; eapply GE1, vstar_app, PR0.
          + eapply kstar_mon; eauto using ext_trans.
        }
      * 
        {
        esplits; eauto.
        right; econs; eauto.
        depdes SU.
        - eapply call_app; eauto using call_app, kstar_kctx, kstar_mon, ext_trans.
        - eapply call_inst; eauto using call_app, kstar_kctx, kstar_mon, ext_trans.
        }
    + 
      rdes PR.
      * 
        esplits; eauto; depdes PR.
        eapply esim_ret; eauto using confrel_mon, vclos_mon, gknow_ant, extpub_trans.
      * depdes PR; rdes RKO; clear ihyp.
        assert (EAB: esim_ py i kma kmb g' sm s'0 true  ea' eb').
        {
          eapply ksim_mon in RKI; [|apply LEB| | |]; eauto.
          destruct RKI; eapply KR; eauto.
        }
        clear RKI. 
        eapply mp; [by eapply KC; econs; eauto using ext_trans|].
        intro X; destruct X.
        specialize (ESIM GK _ _ XTRa XTRb _ _ _ _ CNF0); des.
        eapply esim_main_taus; eauto.
        eapply esim_main_mon; eauto.
        s; i; des; eauto using esim_progress_mon, esim_call_mon, wfo_add_le1.
Qed.

Theorem ctxclo_respectful: resclo (@ctxclo _ _ w).
Proof.
  econs.
  - ii; destruct IN; econs; eauto using kstar_mon.
  - i; destruct PR; eapply key_lemma; eauto.
Qed.

Theorem context_compose: forall (p: esimtyp w),
  compkc (ksim_ (pesim p)) (pesim p) <e= pesim p.
Proof.
  i; eapply pesim_incl_gres.
  eapply gres_incl_clo.
    by apply ctxclo_respectful.
  destruct PR; econs.
  - eauto.
  - eapply kstar_app. 
    rewrite <- (wfo_add_zero_l _ j); eauto.
Qed.

End Context_Closure_Is_Respectful.


(** External Call Elimination Lemma *)


Inductive esim_call_ext {La Lb} {w: world La Lb} kia kib
  gext g si s sync evt mb mb'  ea eb: Prop 
:=
_esim_call i'
    (ST: Lb.(step) evt mb mb')
    (SU: call w s (gext s) (g s) (ksim w i' kia kib g si s s) 
            ea eb)
    (SYN: sync = true)
.
Hint Constructors esim_call_ext.

Inductive esim_ext_ {La Lb} {w: world La Lb} gext (es: esimtyp w) 
  i kia kib g si s sync  ea eb : Prop :=
_esim_ext_
  (ESIM: forall
      (GK: gknow w g)
      cfa cfb (XTRa: extra cfa) (XTRb: extra cfb)
      csa csb ma mb (CNF: configure w g s sync ea eb csa csb cfa cfb ma mb),
    esim_main (esim_progress es i kia kib g si \7/ 
               esim_call_ext kia kib gext g si) 
      kia kib g si s  cfa ma cfb mb).
Hint Constructors esim_ext_.

Definition esim_ext {La Lb} w gext : @esimtyp La Lb w:=
   paco9  (esim_ext_ gext) bote.

Lemma esim_ext_mon La Lb w gext: 
   monotone9  (@esim_ext_ La Lb w gext).
Proof.
  ii; destruct IN; econs; eauto.
  ii; exploit ESIM; eauto using gknow_ant. 
  intro MAIN; des; esplits; eauto.
  eapply esim_main_mon; eauto.
  s; i; des; eauto using esim_progress_mon, esim_call_mon.
Qed.

Hint Resolve esim_ext_mon: paco.

Section External_Call_Elimination_Lemma.

Variables La Lb : LangSpec.
Variable w : world La Lb.

Lemma esim_ext_wfo_le: forall gext i kia kib g si s sync  ea eb
    (ES: esim_ext w gext i kia kib g si s sync  ea eb)
    i' (LE: wfo_le i i'),
  esim_ext w gext i' kia kib g si s sync  ea eb.
Proof.
  i; pfold; punfold ES.
  destruct ES; econs; eauto.
  ii; exploit ESIM; eauto.
  clear ESIM; intro ES; des.
  esplits; eauto.
  eapply esim_main_mon; eauto.
  s; i; des; eauto using esim_progress_mon.
Qed.

Lemma esim_impl_esim_ext: 
  forall i kia kib gext (g: _->vrelf) si s sync  ea eb
    (GEg: forall s  va vb  
            (G: g s  va vb), 
          lsim w g s  va vb \/ gext s  va vb)
    (R: esim w i kia kib g si s sync  ea eb),
  esim_ext w gext i kia kib g si s sync  ea eb.
Proof.
  pcofix CIH.
  i; pfold; apply esim_unfold in R.
  ii; destruct R; econs; eauto.
  ii; exploit ESIM; eauto.
  clear ESIM; intro X.
  depdes X.
  - eauto.
  - eapply esim_step; eauto.
    ii; specialize (EST _ _ ST); des; depdes ES; subst.
    + esplits; eauto.
    + eapply call_union in SU; [|by i; eapply GEg; eauto].
      des; [|by esplits; eauto].
      destruct CNF0; des; esplits; [by eauto| | |by eauto].
      * depdes SU RF VR;
        (left; econs; eauto
        ; right; apply CIH; eauto
        ; eapply pesim_release; [|by eauto|by eauto|by eauto]
        ; eapply context_compose; econs; [eapply CS|]
        ; eauto using vquerya_mon, vqueryb_mon).
      * by econs; eauto; rewrite cplus_emp, <-cplus_assoc.
  - eapply esim_ret; eauto.
Qed.

Lemma esim_ext_impl_esim:
  forall i kia kib gext (g: _->vrelf) si s sync  ea eb
    (R: esim_ext w gext i kia kib g si s sync  ea eb)
    (GEg: gext <g= g),
  esim w i kia kib g si s sync  ea eb.
Proof.
  pcofix CIH.
  i; pfold; punfold R.
  ii; destruct R; econs; eauto.
  ii; exploit ESIM; eauto.
  { eapply gknow_ant; eauto.
    i; eapply gres_incl_id. 
    by left; eapply pesim_mon; eauto. }
  intro X; depdes X; eauto.
  eapply esim_step; eauto; i.
  specialize (EST _ _ ST); des.
  - esplits; eauto.
    left; depdes ES; econs; eauto.
    pclearbot; apply gres_incl_id; eauto.
  - esplits; eauto.
    right; depdes ES; econs; eauto.
    eapply call_mon; eauto.
    i; eapply ksim_mon; eauto.
    i; apply gres_incl_id.
    by left; eapply pesim_mon; eauto.
Qed.

End External_Call_Elimination_Lemma.


Lemma configure_err: forall {La Lb sts} {mc: method_conf La Lb sts}
    s r sync ea eb csa csb cfa cfb ma mb
    (CFG: configure mc s r sync ea eb csa csb cfa cfb ma mb),
  ~ error _ ma /\ ~ error _ mb.
Proof.
  i; inversion CFG; eauto.
Qed.

Lemma pesim_dstep: forall {La Lb} {w: world La Lb} (r: esimtyp w)
    i kia kib g si s  ea eb ma evt ea' ma'
    (RLa: real _ ea ma)
    (DETa: wdet ma)
    (STa: step _ evt ma ma')
    (RLa': real _ ea' ma')
    (X: forall cfb mb (XTR: extra cfb) (RLb: real Lb (eb # cfb) mb), exists eb' mb',
          step _ evt mb mb' /\ real _ (eb' # cfb) mb' /\
          pesim r i kia kib g si s false  ea' eb'),
  pesim r i kia kib g si s false  ea eb.
Proof.
  i; pfold; econs; i; inv CNF; des; subst; noemp.
  esplits; eauto; econs 2; eauto.
    by i; eapply halted_step; [eapply extra_halted|]; rr; eauto.
  i; exploit DETa; [eapply STa| | | |]; eauto; [by rr; esplits; eauto|].
  i; des; subst; hexploit X; eauto; i; des; esplits; eauto.
  - left; econs; eauto using gres_incl_id.
  - econs; eauto; noemp; eauto.
Qed.

Inductive dtsteps {L} : nat -> L.(mach) -> L.(mach) -> Prop :=
| dtsteps_none m
  : dtsteps 0 m m
| dtsteps_some m m' m'' n
    (DET: wdet m)
    (ST: step _ etau m m')
    (ST': dtsteps n m' m'')
  : dtsteps (1 + n) m m''
.

Lemma dtsteps_error: forall L n (m: L.(mach)) c' m'
    (ST: dtsteps n m m')
    (RL: real _ c' m'),
  exists c, real _ c m.
Proof.
  i; ginduction ST; vauto; i; exploit IHST; eauto; i; des.
  apply not_all_not_ex; ii; eapply step_error; eauto.
Qed.

Lemma pesim_dtsteps_aux: forall {n La Lb} {w: world La Lb} (r: esimtyp w)
    i kia kib g si s  ea eb ma ea' ma'
    (RLa: real _ ea ma)
    (STa: dtsteps (S n) ma ma')
    (RLa': real _ ea' ma')
    (X: forall cfb mb (XTR: extra cfb) (RLb: real Lb (eb # cfb) mb),
          pesim r i kia kib g si s false  ea' eb)
    i' (GLE: wfo_le_n (S n) i i'),
  pesim r i' kia kib g si s false  ea eb.
Proof.
  intro n; remember (S n) as n'; ginduction n'; i; inv STa.
  hexploit dtsteps_error; [exact ST'| |]; eauto; i; des.
  i; pfold; econs; i; inv CNF; des; subst; noemp.
  esplits; eauto; econs 2.
  - i; eapply halted_step; [eapply extra_halted|]; rr; eauto.
  - i; exploit DET; [exact ST| | | |]; rr; eauto; destruct n.
    + i; des; subst; hexploit X; eauto; i; des; esplits; eauto.
      * left; econs; eauto using gres_incl_id.
        right; esplits; eauto using wfo_le_n_lt.
      * inv ST'; econs; eauto; noemp; eauto; eapply DET; r; eauto.
    + invc GLE; i; des; subst; hexploit X; eauto; i; des; esplits; eauto.
      * left; econs; eauto.
        { right; esplits; eauto. }
        { apply gres_incl_id; left; eapply IHn'; [|exact X0| | | |]; eauto. }
      * econs; eauto; noemp; eauto.
Qed.

Lemma pesim_dtsteps: forall {n La Lb} {w: world La Lb} (r: esimtyp w)
    i kia kib g si s s'  ea eb ma ea' ma'
    (RLa: real _ ea ma)
    (STa: dtsteps (S n) ma ma')
    (RLa': real _ ea' ma')
    (EXT: ext _ s' s)
    (X: forall cfb mb (XTR: extra cfb) (RLb: real Lb (eb # cfb) mb),
          exists i' eb' mb',
            pesim r i' kia kib g si s' false  ea' eb' /\
            (tc (step _ etau) mb mb' \/
             mb = mb' /\ wfo_le_n (S n) i' i) /\ real _ (eb' # cfb) mb'),
  pesim r i kia kib g si s false  ea eb.
Proof.
  i; pfold; econs; i; inv CNF; des; subst; noemp; inv STa.
  hexploit dtsteps_error; [exact ST'| |]; eauto; i; des.
  hexploit X; eauto; i; des; subst.
  - 
    apply tc_last in X2; des.
    econs 2.
      by i; eapply halted_step; [eapply extra_halted|]; rr; eauto.
    i; exploit DET; [exact ST| | | |]; rr; eauto; destruct n.
    + i; des; subst; esplits; first by eauto.
      * left; econs; eauto using gres_incl_id.
      * econs; eauto; noemp; ss; eauto.
        inv ST'; eapply DET; eauto; rr; eauto.
      * auto.
    + i; des; subst; esplits; first by eauto.
      * left; econs; eauto; apply gres_incl_id.
        left; eapply pesim_dtsteps_aux; [|exact ST'| | |]; eauto.
        instantiate (1 := wfo_add (wfo_nat n) i'); eauto using wfo_le_n_add_S.
      * econs; eauto; noemp; eauto.
      * auto.
  - 
    esplits; eauto; econs 2.
      by i; eapply halted_step; [eapply extra_halted|]; rr; eauto.
    i; exploit DET; [exact ST| | | |]; rr; eauto; destruct n.
    + i; des; subst; esplits; eauto.
      * left; econs; eauto using gres_incl_id.
        right; esplits; eauto using wfo_le_n_lt.
      * econs; eauto; noemp; eauto.
        inv ST'; eapply DET; eauto; rr; eauto.
    + invc X4; i; des; subst; esplits; first by eauto.
      * left; econs; [by right; splits; eauto| |by auto].
        apply gres_incl_id; left; eapply pesim_dtsteps_aux
        ; [|eauto| | |]; eauto.
      * econs; eauto; noemp; eauto.
      * auto.
Qed.


Lemma ksim_done: forall {La Lb w} i kia kib g si sp s 
    (EXT: extpub _ sp si),
  @ksim La Lb w i kia kib g si sp s  kia kib.
Proof.
  i; econs; i; pfold; econs; ii.
  esplits; eauto.
  eapply esim_ret; eauto using extpub_trans.
Qed.
Hint Resolve ksim_done.

Lemma vclos_thunk_lift: forall La Lb (gw: world_glob La Lb) (lw: world_loca gw) sg sl,
  vclos_thunk gw sg <g= vclos_thunk (wlift lw) (sg, sl).
Proof.
  i; inv PR; eauto.
Qed.

Lemma vclos_thunk_unlift: forall La Lb (gw: world_glob La Lb) (lw: world_loca gw) sg sl,
  vclos_thunk (wlift lw) (sg, sl) <g= vclos_thunk gw sg.
Proof.
  i; inv PR; eauto.
Qed.

Lemma vclos_lift: forall La Lb (gw: world_glob La Lb) (lw: world_loca gw) sg sl,
  vclos gw sg <g= vclos (wlift lw) (sg, sl).
Proof.
  i; eapply (vclos_induction PR); i; destruct IH; ss; vauto.
  econs; eapply vclgen_thunk; destruct VR; eauto.
Qed.

Lemma vclos_unlift: forall La Lb (gw: world_glob La Lb) (lw: world_loca gw) sg sl,
  vclos (wlift lw) (sg, sl) <g= vclos gw sg.
Proof.
  i; eapply (vclos_induction PR); i; destruct IH; ss; vauto.
  econs; eapply vclgen_thunk; destruct VR; eauto.
Qed.


Lemma pesim_anti: forall La Lb (w: world La Lb) (r: esimtyp w) i kia kib g si s s'  ea eb
    (SIM: pesim r i kia kib g si s' false  ea eb)
    (EXT: ext _ s' s),
  pesim r i kia kib g si s false  ea eb.
Proof.
  i; punfold SIM; invc SIM.
  pfold; econs; i.
  hexploit ESIM; eauto; first (by inv CNF; eauto); i; des; clear ESIM CNF.
  invc X; esplits; eauto.
  - eapply esim_step; eauto; i.
    hexploit EST; eauto; i; des; esplits; eauto using ext_trans.
  - eapply esim_ret; eauto using ext_trans.
Qed.


Lemma pesim_dance_aux: forall {n La Lb} {w: world La Lb} (r: esimtyp w)
    i kia kib g si s  ea eb ma dma
    (RLa: real _ ea ma)
    (STa: dance ma (S n) dma)
    (X: forall cfb mb (XTR: extra cfb) (RLb: real Lb (eb # cfb) mb)
               cfas ea' (RLa: real _ ea' (dmach_apply dma cfas))
               (XTRa: forall cfa (IN: vIn cfa cfas), extra cfa),
          exists s',
             <<EXT: ext _ s' s>> /\
             <<SIM: pesim r i kia kib g si s' false  ea' eb>>)
    i' (GLE: wfo_le_n (S n) i i'),
  pesim r i' kia kib g si s false  ea eb.
Proof.
  intro n; remember (S n) as n'; ginduction n'; i; depdes STa; vauto.
  i; pfold; econs; i; inv CNF; des; subst; noemp.
  assert (mcplus _ ma cfa ma0) by (rr; eauto).
  hexploit ST; eauto; i; des.
  econs 2; eauto using halted_step.
  i; hexploit DET; eauto; i; des; subst; clear ST1.
  rrdes MM'.
  destruct n.
  - depdes DA.
    hexploit X; eauto.
    instantiate (1 := vcons cfa (vnil _)); s; rewrite <-x; eauto.
    { i; depdes IN; [by auto|by depdes IN].  }
    i; des.
    esplits; auto.
    + left; econs; eauto.
      * right; esplits; eauto using wfo_le_n_lt.
      * eapply gres_incl_id; left; eapply pesim_anti; eauto.
    + econs; eauto; noemp; ss; eauto.
  - invc GLE; i; des; subst.
    esplits; eauto.
    + left; econs; eauto.
      { right; esplits; eauto. }
      { apply gres_incl_id; left; eapply IHn'.
          by eauto. 2: by eauto. by eauto.
          i. eapply X; eauto. instantiate (1:= vcons _ _). eauto.
          i; depdes IN; [by auto|].  eauto.
          eauto. }
    + econs; eauto; noemp; eauto.
Qed.

(** It is crucial that the future state [s'] is quantified inside of
    X, not outside, since it generally depends on the frames. *)
Lemma pesim_dance: forall {n La Lb} {w: world La Lb} (r: esimtyp w)
    i kia kib g si s  ea eb ma dma
    (RLa: real _ ea ma)
    (STa: dance ma (S n) dma)
    (X: forall cfb mb (XTRb: extra cfb) (RLb: real Lb (eb # cfb) mb),
          exists i' eb' mb',
            (forall cfas ea'
                 (RLa: real _ ea' (dmach_apply dma cfas))
                 (XTRa: forall cfa (IN: vIn cfa cfas), extra cfa),
               exists s',
                 <<EXT: ext _ s' s>> /\
                 <<SIM: pesim r i' kia kib g si s' false  ea' eb'>>) /\
            (tc (step _ etau) mb mb' \/
             mb = mb' /\ wfo_le_n (S n) i' i) /\
            real _ (eb' # cfb) mb'),
  pesim r i kia kib g si s false  ea eb.
Proof.
  i; pfold; econs; i; inv CNF; des; subst; noemp.
  depdes STa.
  hexploit X; eauto; i; des; subst.
  - 
    apply tc_last in X1; des.
    assert (mcplus _ ma cfa ma0) by (rr; eauto).
    hexploit ST; eauto; i; des.
    econs 2; eauto using halted_step.
    i; hexploit DET; eauto; i; des; subst; clear ST1.
    rrdes MM'.
    destruct n.
    + depdes DA.
      hexploit (X0 (vcons cfa (vnil _))).
      { by s; rewrite <-x; eauto. }
      { by i; depdes IN; [by auto|by depdes IN]. }
      i; des; esplits; first by eauto.
      * left; econs; eauto using gres_incl_id.
      * econs; eauto; noemp; ss; eauto.
      * auto.
    + i; des; subst; esplits; first by eauto.
      * left; econs; eauto; apply gres_incl_id.
        Require Import ssr.     
        left; eapply pesim_dance_aux; first 1 last; eauto.
        i. eapply X0; eauto. instantiate (1:= vcons _ _). eauto.
        i; depdes IN; [by auto|].  eauto.
        instantiate (1 := wfo_add (wfo_nat n) i'); eauto using wfo_le_n_add_S.
      * econs; eauto; noemp; eauto.
      * auto.
  - 
    assert (mcplus _ ma cfa ma0) by (rr; eauto).
    hexploit ST; eauto; i; des.
    econs 2; eauto using halted_step.
    i; hexploit DET; eauto; i; des; subst; clear ST1.
    rrdes MM'.
    destruct n.
    + depdes DA.
      hexploit (X0 (vcons cfa (vnil _))).
      { by s; rewrite <-x; eauto. }
      { by i; depdes IN; [by auto|by depdes IN]. }
      i; des; subst; esplits; first by eauto.
      * left; econs; eauto using gres_incl_id.
        right; esplits; eauto using wfo_le_n_lt.
      * econs; eauto; noemp; eauto.
      * auto.
    + invc X3; esplits; first by eauto.
      * left; econs; [by right; splits; eauto| |by auto].
        apply gres_incl_id; left; eapply pesim_dance_aux; first 1 last; eauto.
        i. eapply X0; eauto. instantiate (1:= vcons _ _). eauto.
        i; depdes IN; [by auto|].  eauto.
      * econs; eauto; noemp; eauto.
      * auto.
Qed.                            


(** [pesim] is upwards-closed w.r.t. the order index [i]. *)

Lemma pesim_wfo_le: forall La Lb (w: world La Lb) (r: esimtyp w)
    i ka kb g si s sync  (ea: La.(conf)) (eb: Lb.(conf))
    (SIM: pesim r i ka kb g si s sync  ea eb)
    i' (LE: wfo_le i i'),
  pesim r i' ka kb g si s sync  ea eb.
Proof.
  i; punfold SIM; pfold; inv SIM; econs; i.
  hexploit ESIM; eauto; i; des.
  esplits; eauto.
  eapply esim_main_mon; eauto; i; inst; ss; des
  ; eauto using esim_progress_mon, esim_call_mon.
Qed.


(** * Minimal global knowledge for a given world. *)

Inductive gkmin La Lb (w: world La Lb) (g: state w -> vrelf) s : @vrelf La Lb :=
| gkmin_base  va vb (IN: g s  va vb)
  : gkmin w g s  va vb
| gkmin_lsim g'  va vb (IN: lsim w g' s  va vb) (LE: g' <g= gkmin w g)
  : gkmin w g s  va vb

.
Hint Constructors gkmin.


Lemma gkmin_fold: forall La Lb (w: world La Lb) g,
  g \g/ lsim w (gkmin w g)

  <g=
  gkmin w g.
Proof.
  s; i; des; subst; [econs 1|econs 2]; eauto.
Qed.

Lemma gkmin_unfold: forall La Lb (w: world La Lb) g,
  gkmin w g
  <g=
  g \g/ lsim w (gkmin w g)

  .
Proof.
  s; i; inv PR; [by eauto| right].
  - eapply lsim_mon; eauto.

Qed.

Lemma gkmin_gknow: forall La Lb (w: world La Lb) (g : _ -> vrelf)

    (MON: forall s  va vb (R: g s  va vb)
              s' (R: w.(ext) s' s),
            g s'  va vb),
  gknow w (gkmin w g).
Proof.
  i; econs; i; eauto.
  - induction R.
    + econs 1; eauto.
    + econs 2; first (eapply lsim_mon); i; eauto.

Qed.

Lemma gkmin_min: forall La Lb (w: world La Lb) g g'
    (GK: gknow w g')
    (LE: g <g= g'),
  gkmin w g <g= g'.
Proof.
  i; induction PR.
  - eauto.
  - eapply (gk_lsim GK); eauto using lsim_mon.

Qed.

Lemma gkmin_mon: forall La Lb (w: world La Lb)
    g g' (LE: g <g= g'),
  gkmin w g <g= gkmin w g'.
Proof.
  i; ginduction PR; i; eauto.
Qed.

Lemma gkmin_mon': forall La Lb (w: world La Lb)
    g g' (LE: g <g= g')
    (MON: forall s s' (EXT: ext _ s' s), g s <r= g s')
    s s' (EXT: ext _ s' s),
  gkmin w g s <r= gkmin w g' s'.
Proof.
  i; ginduction PR; i.
  - eauto.
  - econs 2; eauto; eapply lsim_mon; eauto.

Qed.

Lemma gkmin_bot_gknow: forall La Lb (w: world La Lb),
  gknow w (gkmin w (fun _ => botr)).
Proof.
  i; econs; i; eauto.
  - induction R.
    + vauto.
    + econs 2; first (eapply lsim_mon); i; eauto.

Qed.


















































Lemma pesim_dance_call_aux: forall {n La Lb} {w: world La Lb} (r: esimtyp w)
    i kia kib g si s  ea eb ma dma
    (RLa: real _ ea ma)
    (STa: dance ma (S n) dma)
    (X: forall cfb mb (XTR: extra cfb) (RLb: real Lb (eb # cfb) mb)
               cfas ea' (RLa: real _ ea' (dmach_apply dma cfas))
               (XTRa: forall cfa (IN: vIn cfa cfas), extra cfa),
          exists s',
             <<EXT: ext _ s' s>> /\
             <<SIM: pesim r i kia kib g si s' false  ea' eb>>)
    i' (GLE: wfo_le_n (S n) i i'),
  pesim r i' kia kib g si s false  ea eb.
Proof.
  intro n; remember (S n) as n'; ginduction n'; i; depdes STa; vauto.

  
  i; pfold; econs; i; inv CNF; des; subst; noemp.
  assert (mcplus _ ma cfa ma0) by (rr; eauto).
  hexploit ST; eauto; i; des.
  econs 2; eauto using halted_step.
  i; hexploit DET; eauto; i; des; subst; clear ST1.
  rrdes MM'.
  destruct n.
  - depdes DA.
    hexploit X; eauto.
    instantiate (1 := vcons cfa (vnil _)); s; rewrite <-x; eauto.
    { i; depdes IN; [by auto|by depdes IN].  }
    i; des.
    esplits; auto.
    + left; econs; eauto.
      * right; esplits; eauto using wfo_le_n_lt.
      * eapply gres_incl_id; left; eapply pesim_anti; eauto.
    + econs; eauto; noemp; ss; eauto.
  - invc GLE; i; des; subst.
    esplits; eauto.
    + left; econs; eauto.
      { right; esplits; eauto. }
      { apply gres_incl_id; left; eapply IHn'.
          by eauto. 2: by eauto. by eauto.
          i. eapply X; eauto. instantiate (1:= vcons _ _). eauto.
          i; depdes IN; [by auto|].  eauto.
          eauto. }
    + econs; eauto; noemp; eauto.
Qed.

Lemma pesim_dance_call: forall {n La Lb} {w: world La Lb} (r: esimtyp w)
    i kia kib g si s  ea eb ma dma
    (RLa: real _ ea ma)
    (STa: dance ma (S n) dma)
    (X: forall cfb mb (XTRb: extra cfb) (RLb: real Lb (eb # cfb) mb),
          exists ea' mb',
            (forall cfas ea''
                 (RLa: real _ ea'' (dmach_apply dma cfas))
                 (XTRa: forall cfa (IN: vIn cfa cfas), extra cfa),
               exists s' j csa eb' csb,
                 <<EXT: ext _ s' s>> /\
                 <<E: ea'' = ea' # csa>> /\
                 <<COa: core ea'>> /\
                 <<ST: tc (step _ etau) mb mb'>> /\
                 <<RLb: real _ (eb' # csb # cfb) mb'>> /\
                 <<COB: core eb'>> /\
                 <<CNF: confrel w g s' csa csb>> /\
                 <<CAL: call w s' (g s') (g s') (ksim_ (pesim r) j kia kib g si s' s')  ea' eb'>>))
    i' (GLE: wfo_le_n (S n) i i'),

  pesim r i' kia kib g si s false  ea eb.
Proof.
  intro n; remember (S n) as n'; ginduction n'; i; [by vauto|]; depdes STa.
  i; pfold; econs; i; inv CNF; des; subst; noemp.
  assert (mcplus _ ma cfa ma0) by (rr; eauto).
  hexploit ST; eauto; i; des.
  econs 2; eauto using halted_step.
  i; hexploit DET; eauto; []; i; des; subst; clear ST1.
  rrdes MM'; destruct n.
  - depdes DA.
    hexploit X; eauto; i; des.
    hexploit (X0 (vcons cfa (vnil _))).
    { by s; rewrite <-x; eauto. }
    { by i; depdes IN; [by auto|by depdes IN]. }
    i; des; subst; rewrite -cplus_assoc in RL.
    apply tc_last in ST1; des.
    esplits; eauto.
    + right; econs; eauto.
      eapply call_mon; eauto using ksim_mon, gres_incl_id.
    + econs; eauto.
  - invc GLE; i; des; subst.
    esplits; eauto.
    + left; econs; eauto.
      { right; splits; eauto. }
      { insts $ eb.
        apply gres_incl_id; left; eapply IHn'
        ; [M| |M| |]; (Mdo by eauto); [by eauto| |by eauto].
        i; hexploit X; eauto; i; des.
        esplits; i; hexploit X0.
        - instantiate (1:= vcons _ _); eauto.
        - i; depdes IN; [by auto|]; eauto.
        - i; des; subst; esplits; eauto.
      }
    + econs; eauto; noemp; eauto.
Qed. 




































Program Definition lwemp La Lb (gw: @world_glob La Lb) : world_loca gw :=
{| lwwfo := gwfo
 ; lwsts := {|state := unit; ext s' s := True; extpub s' s := True |}

 ; lwconf := {|confrel g s ha hb := ha = cemp /\ hb = cemp |}
|}.
Next Obligation.
  eauto using extra_emp.
Qed.

Lemma lwemp_stable: forall La Lb (gw: world_glob La Lb),
  stable (lwemp gw).
Proof.
  ii; esplits; eauto.
Qed.
