(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program.
Require Import countable vgtac basis lang_common lang_spec.
Require Import model_common model_untyped model_properties_untyped gw_untyped mm.
Require Import ssr util opt_transform lang_mid lang_mid_renaming pilsner_common.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



(** Expression transformation hoisting a let-binding out of a definition if possible. *)


(** * Auxiliaries. *)

Lemma perm_help4 G e a b c d:
  wfexp (a :: b :: c :: d :: G) e -> wfexp (d :: a :: b :: c :: G) e.
Proof. 
  rewrite !(cons_app4 _ _ _ _ G). i; eapply wfexp_exchange; eauto.
  eapply Permutation_app; eauto. eauto using Permutation.
Qed.



(** * Definition. *)

Definition hoist_let (y : termvar+contvar) e :=
  match y with
    | inl _ =>
      match e with
        | IEunit | IEnat _ | IEfix _ _ _ _ | IEgen _ _ => true
        | _ => false
      end
    | inr _ =>
      match e with 
        | IEcont _ _ => true
        | _ => false
      end
  end.

Definition hoist (ee: exp) : option exp :=
  match ee with
    | IElet (IEgen k (IElet a (inl y) e) as eg) g e' =>
        if hoist_let (inl y) a && list_disjb (tkvar2var g :: argvars eg) (vars a)
        then Some (IElet a (inl y) (IElet (IEgen k e) g e'))
        else None
    | IElet (IEcont x (IElet a y e) as eg) g e' =>
        if hoist_let y a && list_disjb (tkvar2var g :: argvars eg) (vars a)
        then Some (IElet a y (IElet (IEcont x e) g e'))
        else None
    | IElet (IEfix f x k (IElet a (inl y) e) as eg) g e' =>
        if hoist_let (inl y) a && list_disjb (tkvar2var g :: argvars eg) (vars a)
        then Some (IElet a (inl y) (IElet (IEfix f x k e) g e'))
        else None
    | _ => None
  end.



(** * Syntactic correctness. *)

Lemma hoistP {ea eb}: forall
    (HV: hoist ea = Some eb),
  exists f e' a y e L, 
    <<Ea: ea = IElet (L (IElet a y e)) f e'>> /\
    <<Eb: eb = IElet a y (IElet (L e) f e')>> /\
    <<HL: hoist_let y a = true >> /\
    <<BV: forall x, tkvar2var f = x \/ In x (argvars (L e)) -> In x (vars a) -> False>> /\
    <<CASES: forall NO_DES: unit, 
               (exists bk ty, L = IEgen bk /\ y = inl ty) \/
               (exists bx, L = IEcont bx) \/
               (exists bf bx bk ty, L = IEfix bf bx bk /\ y = inl ty)>>.
Proof.
  i; revert HV. 
  destruct ea; do 2!(try destruct ea1; try discriminate);
  unfold hoist; (try destruct bx1); (try destruct bx0); 
  try destruct (hoist_let _ ea1_1) eqn:HA; try discriminate;
  destruct (list_disjb _ _) eqn:LD; try discriminate;
  s; i; generalize (list_disjb_mem _ _ LD); clear LD;
  s; i; clarify; esplits; eauto 10. 
Qed.

Lemma hoist_wfexp_aux {G ea eb}: forall
    (HV: hoist eb = Some ea)
    (WFb: wfexp G eb)
    (UN: uniq (bvars eb ++ purge_lbl G)),
  <<WFa: wfexp G ea>>.
Proof.
  i. generalize (hoistP HV); ii; rrdes H.
  assert (wfatom G a).
  { rewrite <- (app_nil_l _). eapply (wfatom_STRENGTHEN _ _ (argvars (L e))); last by eauto.
    destruct (CASES ()), f, y; des; clarify; s in WFb; des; try auto;
    try (by eapply wfatom_sublist; first eauto; s; eauto using purge_sublist, sublist_cons_keep);
    try (
    eapply wfatom_exchange; first eapply wfatom_sublist; first eauto; try rewrite cons_app3; 
    try eapply sublist_app; (try by eapply purge_sublist); try (by eapply sublist_refl);
    rewrite app_assoc; eapply Permutation_app; eauto using Permutation).    
  }
  assert (wfexp ([tkvar2var y] ++ argvars (L e) ++ G) e).
  { 
    destruct (CASES ()), f, y; des; clarify; s in WFb; des;
    try (by eapply wfexp_exchange; first eapply wfexp_sublist; try (eassumption);
         try (do !eapply sublist_cons_keep; eapply purge_sublist); eauto 10
        );
    by eapply wfexp_exchange; first eapply wfexp_sublist; try eassumption;
    try (do !eapply sublist_cons_keep; eapply purge_sublist); s; eauto 10; eapply Permutation_cons;
    rewrite !(cons_app3 _ _ _ G); eapply Permutation_app; eauto using Permutation.
  }
  destruct (CASES ()), f,y; s; des; clarify; s; splits; s in WFb; s in H0; des;
     try (by (eauto using wfexp_exchange, perm_swap, perm_help4));
     try (by (rewrite (cons_app _ (_::G)); eapply wfexp_weaken; eauto)).
Qed.

Lemma hoist_wfexp: forall e G e'
    (E: hoist e = Some e')
    (WF: wfexp G e)
    (UN: uniq (bvars e ++ purge_lbl G)),
  <<WF: wfexp G e'>> /\
  <<BV: subuniq (bvars e') (bvars e)>> /\
  <<GT: is_good_thunk e -> is_good_thunk e'>> /\
  <<GE: good_exp e -> good_exp e'>>.
Proof.
  i; generalize (hoistP E); i; des; splits.
  - eapply hoist_wfexp_aux; eauto.
  - apply subuniq_perm.
    destruct (uniq_appD _ _ UN); des.
    clarify; destruct (CASES ()); destruct f, y; des; clarify; s; s in H; s in UN; des;
    rewrite !app_comm_cons !app_assoc;
    eapply Permutation_app_tail;
    rewrite ?app_comm_cons ?app_assoc;
    rewrite (cons_app2 _ _ (bvars e0)); 
    rewrite ?app_assoc;
    try (eapply Permutation_app_tail);
    rewrite 1?(cons_app2 (inl (inr _)));
    rewrite 1?(cons_app2 (inr _));
    try (by eapply Permutation_app_comm);
    rewrite !app_comm_cons !app_assoc;
      eapply Permutation_app_tail;
      s; rewrite (cons_app4); eapply Permutation_app_comm.
  - clarify; s; i.
  - clarify; s; i.
Qed.



(** * Semantic correctness. *)

Lemma real_midP {h} {env} {e} {cf} {mb}:
  real mid_lsi (((h, Just (env, e)) : conf mid_lsi) # cf) mb
  -> exists hf, cf = (hf, Nothing) /\ mb = (h ## hf, (env, e)) /\ pmvalid (h ## hf) /\ pmfinite (h ## hf).
Proof. i. rrdes H. destruct cf. clarify. s in Ec. destruct e0; try discriminate. clarify. eauto. Qed.

Lemma hoist_letPl:
  forall {y} {e} (HL : hoist_let (inl y) e) (enva : env), exists va, eval enva e = Some va.
Proof. i. destruct e; try discriminate; esplits; s; eauto. Qed.

Lemma hoist_letPr:
  forall {ky} {e} (HL : hoist_let (inr ky) e), exists k x, e = IEcont k x.
Proof. i. destruct e; try discriminate; esplits; eauto. Qed.


Lemma RV_app 
        {X : Type} {f : X -> var}
        {G : list var} G' {enva} enva' {enva''} {envb} envb' {envb''}
        {P : vrel}:
  forall (RV' : forall x,
                 In (f x) (G') ->
                 exists va vb : lang_spec.val mid_lsi,
                   << INa : lookup (f x) (enva') = Some va >> /\
                   << INb : lookup (f x) (envb') = Some vb >> /\
                   << PR : P va vb >>)
         (DISJ : forall x (INxG : In (f x) G') (INx : (In (f x) (map fst enva) \/ In (f x) (map fst envb))), False) 
         (RV : forall x (ING : In (f x) G),
            exists va vb : lang_spec.val mid_lsi,
              << INa : lookup (f x) (enva) = Some va >> /\
              << INb : lookup (f x) (envb) = Some vb >> /\
              << PR : P va vb >>)
         (ENVa : enva'' = enva ++ enva')
         (ENVb : envb'' = envb ++ envb'),
  forall x,
    In (f x) (G ++ G') ->
    exists va vb : lang_spec.val mid_lsi,
      << INa : lookup (f x) (enva'') = Some va >> /\
      << INb : lookup (f x) (envb'') = Some vb >> /\
      << PR : P va vb >>.
Proof.
  i. subst. apply in_app_or in H; des.
  - destruct (RV x) as [va [vb ]]; auto; des.
    exists va vb. by rewrite !list_assoc_app INa INb.
  - destruct (RV' x) as [va [vb ]]; auto; des.
    exists va vb. rewrite !list_assoc_app INa INb.
    destruct (lookup (f x) enva) as [va'|] eqn:HA;
    destruct (lookup (f x) envb) as [vb'|] eqn:HB;
    i; eauto; exfalso; eapply DISJ; eauto.
    + apply list_assoc_someD, (in_map fst) in HA; auto.
    + apply list_assoc_someD, (in_map fst) in HA; auto.
    + apply list_assoc_someD, (in_map fst) in HB; auto.
Qed.


Lemma ren_contvar_id: forall x,
  rename_contvar ren_id x = x.
Proof.
    by destruct x as [[]|].
Qed.                            

Lemma hoist_psim lw G ea eb : forall
    (HV: hoist eb = Some ea)
    (WFb: wfexp G eb)
    (UN: uniq (bvars eb ++ purge_lbl G))
    (SLW: stable lw),
  mm_psim lw G ren_id ea eb.
Proof.
  i. generalize (hoistP HV); generalize (hoist_wfexp_aux HV WFb UN); intro WFa; i; rrdes H.
  clarify. clear HV.
  rr; splits; try by destruct (CASES ()); des.
  exists 0; i. eapply pesim_acquire; i; eapply pesim_real; i.
  destruct hsa as [? eea], hsb as [? eeb]. 
  destruct (real_midP RLa) as [ha []], (real_midP RLb) as [hb []];
  clarify; s; s in RLa; s in RLb; noemp; clear RLa RLb; des; clarify.
  
  assert (INyG : In (tkvar2var y) G -> False).
  { i; s; destruct (CASES ()), f, y; des; clarify; eapply (uniq_appD _ _ UN); 
    try eassumption; s; eauto 10; eapply (purge_lbl_tkvar2var _ _ H). 
  }
  assert (IngG : In (tkvar2var f) G -> False).
  { i; s; destruct (CASES ()), f, y; des; clarify; eapply (uniq_appD _ _ UN);
    try eapply (purge_lbl_tkvar2var _ _ H); s; eauto 10.
  }
  assert (INbvG : forall x, tkvar2var f = x \/ In x (argvars (L e)) -> In x G -> False).
  { destruct (uniq_appD _ _ UN) as [_ [_ F]];
    i; s; destruct (CASES ()), f, y; des; clarify; s in H; s in F; des; clarify;
    try (eapply F; last (by eapply (purge_lbl_tkvar2var (inr _) G H0)); first by eauto);
    try (eapply F; last (by eapply (purge_lbl_tkvar2var (inl _) G H0)); first by eauto).
  }
  assert (UNV : uniq ((tkvar2var y) :: (tkvar2var f) :: argvars (L e))).
  { destruct (CASES ()), f, y; des; clarify; s; s in UN; des; splits; i; des; eauto 10. }
    
  destruct (LEM (exists pa, reduce_ok etau ha enva (IElet (L (IElet a y e)) f e') ha (pa::enva) e')) 
           as [STEPga|NOSTEP]; last first.
  { i; pfold; econs 1; i. inv CNF. destruct (real_midP MA) as [ha'], (real_midP MB) as [hb']; des.
    noemp; clarify; noemp; clarify. 
    econs.
    + econs 1. eapply IRerr; auto; i. 
      inv FH; eapply NOSTEP; destruct (CASES ()); des; clarify; try (inv EVAL); eexists; econs; by econs.
    + eauto.
  }

  destruct (LEM (exists pb, reduce_ok etau hb envb (IElet (L (IElet a y e)) f e') hb (pb::envb) e'))
  as [STEPgb|NOSTEP]; last first.
  { i. pfold. econs 1. i.
    destruct XTRa as [Xha XTRa], XTRb as [Xhb XTRb]. subst.
    inv CNF. destruct (real_midP MA) as [ha'], (real_midP MB) as [hb']; des. 
    noemp; clarify; noemp; clarify.
    rrdes STEPga; inv STEPga; destruct (CASES ()); des; noemp; clarify; noemp; clarify
    ; by econs; first econs 3; inst_pairs;
         try (econs 1; econs 2; i; first by auto); try (by inv FH; eapply NOSTEP; by eauto).
  }

  destruct (@mm_psim_refl lw e' (tkvar2var f :: G)) as [_ [_ [i' PSe]]]; auto.
  { s in WFb. destruct f; des; by auto. }  

  eapply pesim_stepb; i.
  destruct (real_midP RL) as [h]; des; clarify; s; s in RL; noemp.
  
  assert (reduce_ok etau (hb ## h) envb (IElet (L (IElet a y e)) f e') (hb ## h) ((tkvar2var f, IVclo envb (L (IElet a y e)))::envb) e').
  { i. des. inv STEPgb; destruct (CASES ()); des; clarify; clean; esplits; inst_pairs; (econs; econs; by econs). }
  
  eexists (hb, Just ((tkvar2var f, _)::envb, e')). 
  eexists (hb ## h, ((tkvar2var f, IVclo envb (L (IElet a y e)))::envb, e')). 
  esplits; first by econs; econs.
  right. s. esplits.
  { econs. esplits; try reflexivity; auto. }
  
  assert (TCa1 : exists vya,
                  reduce_ok etau ha enva (IElet a y (IElet (L e) f e'))
                                 ha ((tkvar2var y, vya) :: enva) (IElet (L e) f e')
         ).
  { destruct y as [ty|ky]; 
      [destruct (hoist_letPl HL enva) as [va Eva]|destruct (hoist_letPr HL) as [kx ke]]; 
      destruct (CASES ()); invc STEPga; des; clarify; try (s in EVAL; clarify); econs; try by econs; eassumption.
  } des.
  
  assert (ASa: mm_asim lw G ren_id a a).
  { s in WFa; destruct y; des; eapply mm_asim_refl; auto. }
  
  assert (KSa: mm_ksim lw G ren_id a a).
  { s in WFa; destruct y; des; eapply mm_ksim_refl; auto. }
  
  
  assert (KSa_purge : forall (NC : forall x, L <> IEcont x), mm_ksim lw (purge G) ren_id a a).
  { ii; eapply mm_ksim_refl; eauto. 
    destruct (CASES ()), f, y; des; clarify; s in WFb; des; clarify; s; (try by exfalso; eapply NC; auto);
    try (by rewrite (cons_app _ (purge G)); eapply wfexp_strengthen; s; eauto; eapply BV; s; eauto);
    by rewrite (cons_app _ (purge G)); 
      eapply wfexp_strengthen; first eapply wfexp_strengthen; first eapply wfexp_strengthen; 
      first (s; eassumption); i; s in INbvG; s in BV; eauto 10.
  }
     
  assert (Evyb_cont : 
          forall x h envb'
                 (DISJ: forall x, In x (map fst envb') -> In x G -> False),
            L = IEcont x ->
            exists vyb, 
              << IRVy : reduce_ok etau h (envb' ++ envb) (IElet a y e) h ((tkvar2var y, vyb)::(envb' ++ envb)) e >>
              /\ << VCLvy : forall y', y = inl y' -> vclos (wlift lw) s (g s) vya vyb >>
              /\ << KSvy : forall y', y = inr y' -> ksim (wlift lw) (wfo_nat i) kia kib g si s s vya vyb >>).
  {
    ii; clarify. 
    ii. destruct y as [ty|ky].
    - edestruct (hoist_letPl HL (envb'++envb)).
      edestruct ASa; des; try exact H0; try eassumption.
      + ii. instantiate (3 := enva).
        rewrite list_assoc_app.
        destruct (lookup _ envb') eqn:L.
        * exfalso; eapply DISJ; eauto.
          apply list_assoc_someD, (in_map fst) in L; auto.
        * eauto.
      + invc TCa1. exists x0; splits; i; clarify. by econs.
    - edestruct (hoist_letPr HL); des.
      edestruct (KSa g s si i (wfo_nat i) kia kib); des; try exact H0.
      + eassumption.
      + eassumption.
      + ii. instantiate (1 := envb' ++ envb).
        instantiate (1 := enva).
        rewrite list_assoc_app.
        destruct (lookup _ envb') eqn:L.
        * exfalso; eapply DISJ; eauto.
          apply list_assoc_someD, (in_map fst) in L; auto.
        * destruct (RV _ IN); des; esplits; eauto.
      + ii.
        rewrite list_assoc_app.
        destruct (lookup _ envb') eqn:L.
        * exfalso; eapply DISJ; eauto.
          apply list_assoc_someD, (in_map fst) in L; auto.
        * destruct (RK _ IN); des; esplits; eauto. 
      + invc TCa1.
        esplits; i; clarify.
        * by econs.
        * exact SIM.
  }
  
  assert (Evyb : 
          forall s' envb'
                 (DISJ: forall x, In x (map fst envb') -> In x G -> False)
                 (LE : forall va vb, g s va vb -> g s' va vb)
                 (EXTs : ext (wlift lw) s' s)
                 (NC : (forall x, L <> IEcont x)),
          exists vyb, << IRVy : forall h, reduce_ok etau h (envb' ++ envb) (IElet a y e) h ((tkvar2var y, vyb)::(envb' ++ envb)) e >>
          /\ << VCLvy : forall y', y = inl y' -> vclos (wlift lw) s' (g s') vya vyb >>).
  { 
    ii. destruct y as [ty|ky].
    - edestruct (hoist_letPl HL (envb'++envb)).
      edestruct (ASa g s enva (envb'++envb)); des; try exact H1; try eassumption.
      + ii. rewrite list_assoc_app.
        destruct (lookup _ envb') eqn:E; eauto.
        exfalso; eapply DISJ; eauto.
        apply list_assoc_someD, (in_map fst) in E; auto.
      + invc TCa1. exists x; splits; i; clarify. by econs.
        eapply vclos_mon; eauto.
    - edestruct (hoist_letPr HL); des.
      invc TCa1. esplits; i; clarify. by econs.
  }
  
  assert (TCa2 :  
          reduce_ok etau ha ((tkvar2var y, vya) :: enva) (IElet (L e) f e')
                    ha ((tkvar2var f, IVclo ((tkvar2var y, vya)::enva) (L e)) :: ((tkvar2var y, vya)::enva)) e'
         ).
  { destruct y as [ty|ky]; 
      [destruct (hoist_letPl HL enva) as [va Eva]|destruct (hoist_letPr HL) as [kx ke]]; 
      destruct (CASES ()); invc STEPga; des; clarify; try (s in EVAL; clarify); econs; try by econs; eassumption.
  } des.
  
  assert (Hcv: forall i kf, 
            f = inr kf ->
            ksim (wlift lw) (i) kia kib g si s s
                 (IVclo ((tkvar2var y, vya) :: enva) (L e))
                 (IVclo envb (L (IElet a y e)))).
  {
    i; clarify.
    destruct (CASES ()); des; clarify; try (by invc TCa2).
    econs; i. invc RETa RETb. 
    eapply pesim_gknow; i; eapply pesim_gknow_; i.
    destruct (@mm_psim_refl lw e ([tkvar2var y; inl (inr bx)] ++ G)) as [_ [_ [i'' PSe0]]]; try (by s; auto).
    { revert WFb. try destruct y; s; i; des; auto. }
    eapply pesim_acquire; i. invc HR0. des. noemp.
    eapply pesim_real; i.
    apply real_midP in RLa. eapply real_midP in RLb. des; clarify.
    dup WHX; apply mm_confrel_shape in WHX0; des; clarify.
    destruct hby, hay; destruct e0, e1; clarify. s. noemp. simpl in RLa, RLb. clarify.
    eapply pesim_dtsteps.
    - rr; esplits; eauto.
    - econs 2; inst_pairs; first by eapply mid_wdet. 
      + econs; econs; eassumption.
      + econs 1.
    - rr; esplits; eauto.
    - auto.
    - i. apply real_midP in RLb; des; clarify. invc XTR0; clarify.
      edestruct (Evyb_cont bx); try (by s; eauto); des.
      { s; i; des; clarify; eauto.
        eapply INbvG; last eassumption.
        instantiate (1:=[(inl (inr bx), vb)]) in H0.
        s in H0; des; s; auto.
      }
      esplits; inst_pairs; last 2 first.
      + left. econs 2. 
        * econs; inst_pairs. econs; econs; eassumption.
        * econs 1; inst_pairs; econs; eauto. 
      + rr; esplits; s; try erewrite eoplus_nothing; eauto.
      +
      rewrite (mid_csplit (ha' ## h1)) (mid_csplit (hb' ## h0)). 
      eapply pesim_release; eauto; try (by rr; esplits; eauto; s; eauto).
      eapply pesim_mon; first (eapply pesim_gknow; i; eapply PSe0; eauto); last by ii; auto.
      * eauto using extpub_trans.
      * i; rewrite ren_itvar_id. revert x1 IN. eapply RV_app; try by rewrite cons_app2.
        { i. destruct (RV x1 H0); des.
          rewrite ->ren_itvar_id in *.
          esplits; eauto. eapply vclos_mon; eauto; i; eapply gk_mon; eauto. }
        { s; i; des; clarify; eapply INbvG; s; try (by auto); rewrite <- INx in INxG;
           rewrite cons_app2 in UNV; exfalso; eapply (uniq_appD _ _ UN); try (eassumption);
           try eapply (purge_lbl_tkvar2var _ _ INxG); destruct y; s; eauto.
        }
        { s In; destruct y; s tkvar2var; i; des; clarify.
          - rewrite list_assoc_cons_neq; first rewrite 2!list_assoc_cons_eq;
            last by ( s in UNV; des; rr; i; clarify; auto).
            esplits; eauto. eapply vclos_mon; first eauto; eauto. i; eapply gk_mon; try eassumption; eauto.
          - rewrite list_assoc_cons_eq list_assoc_cons_neq; first rewrite list_assoc_cons_eq;
            last by ( s in UNV; des; rr; i; clarify; auto).
            esplits; eauto.
          - rewrite list_assoc_cons_eq list_assoc_cons_neq; first rewrite list_assoc_cons_eq;
            last by ( s in UNV; des; rr; i; clarify; auto).
            esplits; eauto.
        }
    * eapply RV_app; try by rewrite cons_app2.
        { i; destruct (RK x1 H0); des. esplits; eauto. eapply ksim_mon; eauto. }
        { s; i; des; clarify; eapply INbvG; s; try (by auto); rewrite <- INx in INxG;
           rewrite cons_app2 in UNV; exfalso; eapply (uniq_appD _ _ UN); try (eassumption);
           try eapply (purge_lbl_tkvar2var _ _ INxG); destruct y; s; eauto.
        } 
        { s In; destruct y; s tkvar2var; i; des; clarify.
          - rewrite list_assoc_cons_neq; first rewrite !list_assoc_cons_eq;
            try by ( s in UNV; des; rr; i; clarify; auto).
            esplits; eauto. eapply ksim_mon; first eapply KSvy; eauto.
        }
  }
  
  assert (Htv: forall tf, 
            f = inl tf ->
            vclos (wlift lw) s (g s) (IVclo ((tkvar2var y, vya) :: enva) (L e))
                  (IVclo envb (L (IElet a y e)))
  ).
  { i; clarify. 
    econs.
    eapply vclgen_thunk. destruct (CASES ()); des; clarify; last 2 first.
    - by invc TCa2.
    - eapply vclf_fix; try by econs; eapply gk_lsim; eauto; econs.
      eapply gk_lsim; eauto. 

      destruct (@mm_psim_refl lw e (([inl (inr ty);inr bk;inl (inr bx);inl (inr bf)]) ++ purge G))
        as [_ [_ [i'' PSe0]]]; try (by s; auto).
      { revert WFb. try destruct y; s; i; des; auto. }

      evar (i_: wfo_set (wlift lw)).
      econs; first by (econs; econs; intro HF; invc HF).
      instantiate (1:=i_).
      pcofix CIH; i.
      unfold i_ in *. clear i_.
      invc STK; last by invc RK0 INSTa INSTb; invc RF; invc CA CB VR.
      invc RK0 APPa APPb RF; invc CA CB VR. 
      eapply pesim_gknow; i; eapply pesim_gknow_; i.
      eapply pesim_acquire; i. invc HR0. rename x into hc.

      edestruct (Evyb s'); try (by s; eauto); des.
      { s; i; des; clarify; eauto.
        eapply INbvG; last eassumption.
        instantiate (1:=[(inr bk1, kb); (inl (inr bx1), vb);
        (inl (inr bf1), IVclo env0 (IEfix bf1 bx1 bk1 (IElet a (inl ty) e0)))]) in H0.
        s in H0; des; s; auto.
      }
      { i; eapply gk_mon; eauto. }
      { ii; clarify. congruence. }

      eapply pesim_stepb; i; rrdes RL0; clarify.
      destruct hbx, hby, cfb.
      invc XTR0. noemp. destruct e, e1; simpl in Ec; clarify.
      noemp. rewrite cons_app3.

      eexists ((h1 ## h2, Just (_,e0))); esplits; inst_pairs; first by econs; econs; eauto.
      right; split; first by rr; esplits; s; try (by s; auto).
      rewrite (mid_csplit (h1 ## h2)). 
      eapply pesim_release; eauto; try (by rr; esplits; eauto; s; eauto).
      eapply pesim_mon; first (eapply pesim_gknow; i; eapply PSe0; eauto); last by ii; auto.
      { i; rewrite ren_itvar_id. revert x1 IN.
        eapply (RV_app (purge G) enva env0); try (by rewrite cons_app4; auto).
        - i; destruct (RV _ (purge_inD _ _ H0)); des.
          rewrite ->ren_itvar_id in *. esplits; eauto.
          eapply vclos_mon; first eassumption; eauto; i; eapply gk_mon; eauto.
        - destruct (uniq_appD _ _ UN) as [_ [_ F]]; s; i; des; clarify; try (rewrite <- INx in INxG; clear INx);
          try apply purge_inDr in INxG; auto;
          try apply purge_inD in INxG;
          try (eapply F; last (by eapply (purge_lbl_tkvar2var (inr _) G INxG)); first by s; eauto);
          try (eapply F; last (by eapply (purge_lbl_tkvar2var (inl _) G INxG)); first by s; eauto).
        - s In; i; s in ING; des; clarify; first (des; clarify).
        + rewrite list_assoc_cons_neq; first rewrite list_assoc_cons_neq;
          first rewrite list_assoc_cons_eq list_assoc_cons_neq;
          first rewrite list_assoc_cons_eq;
          try by ( s in UNV; des; rr; i; clarify; auto).
          esplits; eauto.
          eapply vclos_mon; first eauto; eauto. 
        + rewrite list_assoc_cons_neq; first rewrite list_assoc_cons_eq 2?list_assoc_cons_neq;
          first rewrite list_assoc_cons_eq;
          try by ( s in UNV; des; rr; i; clarify; auto).
          esplits; eauto.
        + rewrite 2?list_assoc_cons_neq;
          first rewrite list_assoc_cons_eq 3?list_assoc_cons_neq;
          first rewrite list_assoc_cons_eq;
          try by ( s in UNV; des; rr; i; clarify; auto).
          esplits; eauto.
          econs. eapply vclgen_thunk. eapply vclf_fix; try by econs.
          eapply gk_lsim; try exact GK1. eauto. econs; i; first by (econs; econs; intro HF; invc HF).
          eapply gres_incl_id.
          inv STK; last by invc RK0 INSTa INSTb; invc RF; invc CA CB VR.
          invc RK0 APPa APPb RF; last by invc CA CB VR. 
          invc CA CB VR.
          right. eapply CIH; esplits; eauto using ext_trans. 
      }
      {
        eapply RV_app; try (by rewrite cons_app4); try (by i; exfalso; eauto using purge_inDr).
        - s In; i; des; clarify; first try (s in ING); clarify.
          + rewrite list_assoc_cons_eq list_assoc_cons_neq ?list_assoc_cons_eq; try (ii; discriminate).
            * esplits; eauto.
      }
    - eapply vclf_gen; try by econs; eapply gk_lsim; eauto; econs.
      eapply gk_lsim; eauto. econs; first by (econs 2; econs; intro HF; invc HF).
      i. invc STK; first by invc RK0 APPa APPb; invc RF; invc CA CB VR.
      invc RK0 INSTa INSTb RF; invc CA CB VR.
      eapply pesim_gknow; i; eapply pesim_gknow_; i.
      destruct (@mm_psim_refl lw e0 (([inl (inr ty);inr bk1]) ++ purge G)) as [_ [_ [i'' PSe0]]]; auto.
      { revert WFb. try destruct y; s; i; des; auto. }
      eapply pesim_acquire; i. invc HR0. rename x into hc.

      edestruct (Evyb s'); try (by s; eauto); des.
      { s; i; des; clarify; eauto.
        eapply INbvG; last eassumption.
        instantiate (1:=[(inr bk1, kb)]) in H0.
        s in H0; des; s; auto.
      }
      { i; eapply gk_mon; eauto. }
      { ii; clarify; congruence. }

      eapply pesim_stepb; i; rrdes RL0; clarify.
      destruct hbx, hby, cfb.
      invc XTR0. noemp. destruct e, e1; simpl in Ec; clarify.
      noemp. rewrite cons_app.

      eexists ((h1 ## h2, Just (_,e0))); esplits; inst_pairs; first by econs; econs; eauto. 
      right; split; first by rr; esplits; s; try auto.
      rewrite (mid_csplit (h1 ## h2)). 
      eapply pesim_release; eauto; try (by rr; esplits; eauto).
      eapply pesim_mon; first (eapply pesim_gknow; i; eapply PSe0; eauto); last ii; clarify.
      { i; rewrite ren_itvar_id. revert x1 IN.
        eapply (RV_app (purge G) enva env0); try rewrite cons_app2; auto.
        - i; destruct (RV _ (purge_inD _ _ H0)); des.
          rewrite ->ren_itvar_id in *. esplits; eauto.
          eapply vclos_mon; first eassumption; eauto; i; eapply gk_mon; eauto.
        - destruct (uniq_appD _ _ UN) as [_ [_ F]]; s; i; des; clarify; try (rewrite <- INx in INxG; clear INx);
          try apply purge_inDr in INxG; auto;
          try apply purge_inD in INxG;
          try (eapply F; last (by eapply (purge_lbl_tkvar2var (inr _) G INxG)); first by s; eauto);
          try (eapply F; last (by eapply (purge_lbl_tkvar2var (inl _) G INxG)); first by s; eauto).
        - s In; i; s in ING; des; clarify; first (des; clarify).
        + rewrite list_assoc_cons_neq 1?list_assoc_cons_eq
                  1?list_assoc_cons_eq;
          try by ( s in UNV; des; rr; i; clarify; auto).
          esplits; eauto.
          eapply vclos_mon; first eauto; eauto. 
      }
      {
        eapply RV_app; try (by rewrite cons_app2); try (by i; exfalso; eauto using purge_inDr).
        s In; i; des; clarify; first try (s in ING); clarify.
        + rewrite list_assoc_cons_eq list_assoc_cons_neq ?list_assoc_cons_eq; try (ii; discriminate).
          * esplits; eauto.
      }
  }
  
  eapply pesim_dtsteps.
  - rr; esplits; try reflexivity; by auto.
  - econs; first by eapply mid_wdet.
    * inst_pairs. econs; eauto.
    * econs 2; first by eapply mid_wdet.
      inst_pairs. econs; eauto.
      econs 1.
  - rr; esplits; try reflexivity; by auto.
  - by eauto.
  - i. eexists. exists ((hb, Just ((tkvar2var f, IVclo envb (L (IElet a y e)))::envb, e'))). 
    esplits; last 1 first.
    + invc XTR0 RLb; des; clarify. rr; esplits; eauto.
    + rewrite (mid_csplit ha) (mid_csplit hb).
      eapply pesim_release; i; try auto; try by rr; esplits; eauto.
      eapply PSe; auto; last first.
      * s In; i; des.
        { 
rewrite ren_contvar_id.
          rewrite <- IN, !list_assoc_cons_eq. 
          invc H; destruct (CASES ()); des; clarify; clean.
          esplits; eauto.
        }
        { 
          rewrite list_assoc_cons_neq. rewrite list_assoc_cons_neq.
          rewrite list_assoc_cons_neq. 
          - by eauto.
          - destruct f; s;ii; clarify.
          - destruct f,y; s;ii; clarify.
          - destruct f; s;ii; clarify.
        }
      * intros [lb|x]; s In; [|i; des; clarify].
        { 
          s In; i; des; first by destruct f. 
          rewrite !list_assoc_cons_neq; try (destruct f,y; s; congruence).
          eauto using RV.
        }
        {  
          rewrite ->ren_itvar_id, <- IN. rewrite !list_assoc_cons_eq.
          esplits; eauto. destruct f; clarify. eapply Htv; eauto. 
        }
        { 
          rewrite list_assoc_cons_neq. rewrite list_assoc_cons_neq.
          rewrite list_assoc_cons_neq. 
          - eauto.
          - destruct f; s;ii; clarify.
          - destruct f,y; s;ii; clarify.
          - destruct f; s;ii; clarify.
        }
    +  right; splits; eauto using wfo_le_n_add_S.
Grab Existential Variables. 
  exact i. exact wfo_zero. exact wfo_zero. exact i.
Qed.



Definition hoist1_ann: annT.
  refine {| _annT := unit
          ; perform u := hoist
         |}.
  - ii; eauto using hoist_wfexp.
  - ii; econs; eauto using hoist_psim.
Defined.

Definition hoist_ann: annT :=
  annT_list hoist1_ann.

Definition max_iterations := 42.

Definition analyze: exp -> aexp hoist_ann :=
  annotate_with (repeat tt max_iterations).

Definition hoist_mod: modtf :=
  fun _ => transform_mod analyze.

Theorem hoist_wfmod: forall FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m),
  mid_wfmod FI (hoist_mod FI m) FO /\
  uniqvar (hoist_mod FI m).
Proof.
  i; eapply transform_wfmod; eauto using annotate_with_erase.
Qed.

Theorem hoist_msim: forall FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m),
  rtc (msim mm_gw FI FO) (hoist_mod FI m) m.
Proof.
  i; eapply transform_msim; eauto using annotate_with_erase.
Qed.  

Theorem hoist_correct:
  <<WF: modtf_wf hoist_mod>> /\
  <<SIM: modtf_sim hoist_mod>>.
Proof.
  split; ii; eauto using hoist_wfmod, hoist_msim.
Qed.
