(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Relations ClassicalChoice Wellfounded.
Require Import paco2 vgtac basis countable.
Require Import lang_common lang_spec types.
Require Import model_common model model_properties.
Require Import gw lang_tgt lang_src_lsi.
Require Import util ssr.

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".


(** * Proof of adequacy

    Adequacy is the ultimate correctness property.  It states that if
    a source program is related to a target program and they are
    complete (closed), then the target program actually refines the
    source program.  We prove this for the main model, not
    generically.

    The development is done for the generic model, except for the
    final theorem about [msim], which is proved for the main model.
*)


Inductive obs_event :=
| oein (n : nat)
| oeout (n: nat)
.

Coercion o2e (evt: obs_event) : event :=
  match evt with
  | oein n => ein n
  | oeout n => eout n
  end.

CoInductive observation : Set :=
| obs_done
| obs_inftau
| obs_evt (evt: obs_event) (obs: observation)
.

Inductive behmatch {L} behave : L.(mach) -> observation -> Prop :=
| beh_err m obs
    (ERR: L.(error) m)
  : behmatch behave m obs
| beh_done m
    (TM: L.(halted) m)
  : behmatch behave m obs_done
| beh_inftau m m'
    (ST: L.(step) etau m m')
    (INF: behave m' obs_inftau : Prop)
  : behmatch behave m obs_inftau
| beh_evt m m' (evt: obs_event) obs
    (ST: L.(step) evt m m')
    (INF: behave m' obs : Prop)
  : behmatch behave m (obs_evt evt obs)
.
Hint Constructors behmatch.

Inductive behave_ {L} behave m obs : Prop :=
behave_intro m'
  (MAT: behmatch behave m' obs)
  (TAU: rtc (L.(step) etau) m m').
Hint Constructors behave_.

Definition behave {L} : _ -> _ -> _ -> Prop := paco2 (@behave_ L).

Lemma behave_mon L: monotone2 (@behave_ L).
Proof.
  ii; destruct IN; destruct MAT; eauto.
Qed.
Hint Resolve behave_mon: paco.

Lemma behave_taus L: forall r mb mb' obs
    (BEH: behave r mb obs)
    (TAU: rtc (L.(step) etau) mb' mb),
  behave r mb' obs.
Proof.
  i; pfold; punfold BEH; destruct BEH; eauto using rt_trans.
Qed.

Lemma behave_inftau_unfold L: forall (m: L.(mach))
    (R: behave bot2 m obs_inftau),
  behmatch (behave bot2) m obs_inftau.
Proof.
  i; punfold R; depdes R; apply clos_rt_rt1n in TAU; destruct TAU.
  - destruct MAT; pclearbot; eauto.
  - eapply beh_inftau; eauto. 
    pfold; econs; eauto using clos_rt1n_rt. 
Qed.

Lemma adequacy_esim La Lb: forall w i kia kib g si s sync ty ea eb
    (GK: gknow w g)
    (CONS: forall s (fty: flextyp) va vb (TH: is_thunktyp fty)
           (G: g s fty va vb), lsim w g s fty va vb)
    (RKa: forall
            s (EXTP: w.(extpub) s si)
            va ea (RET: w.(cquerya) s (ret va kia) ea)
            ha ma (MA: La.(real) (ea # ha) ma),
          La.(halted) ma)
    (RKb: forall
            s (EXTP: w.(extpub) s si)
            vb eb (RET: w.(cqueryb) s (ret vb kib) eb)
            hb mb (MB: Lb.(real) (eb # hb) mb),
          Lb.(halted) mb)
    (SIM: esim w i kia kib g si s sync ty ea eb)
    cfa cfb (XTRa: extra cfa) (XTRb: extra cfb)
    csa csb ma mb (CFG: configure w g s sync ea eb csa csb cfa cfb ma mb)
    obs (OBS: behave bot2 ma obs),
  behave bot2 mb obs.
Proof.
  intros until csa; revert csa.
  revert i s sync ty ea eb SIM; pcofix CIH.
  i; eapply esim_impl_esim_ext with (gext := fun _ => botr) in SIM; eauto.
  i; punfold SIM.
  destruct SIM.
  exploit ESIM; eauto.
  clear ESIM; intro SIM; des.

  punfold OBS; destruct OBS.
  move TAU after i; revert_until TAU.
  apply clos_rt_rt1n in TAU.
  induction TAU as [ma | ma ma' ma'' TAU].
  {
    intro i; revert ma; assert (WF := wfo_wf _ i).
    induction WF as [i ACC IH].
    i; destruct SIM; [by eauto| |].
    { i; depdes MAT.
      - destruct CFG; exfalso; eapply ERR; eauto.
      - exfalso; eauto.
      - specialize (EST _ _ ST); des; ss; [|by depdes ES SU RF VR].
        depdes ES; pclearbot; des.
        + pfold; econs; [|edone].
          eapply beh_inftau; eauto.
          right; eapply CIH; try apply CNF; eauto.
          by eapply esim_ext_impl_esim; eauto.
        + subst; apply tc_last in LT; des.
          eapply behave_taus; [|edone].
          punfold ES; destruct ES.
          exploit ESIM; eauto.
          clear ESIM; intro ES; des.
          eapply IH with (s:=s'); eauto.
          * apply behave_inftau_unfold in INF.
            depdes INF; pclearbot; eauto.
          * eapply esim_main_mon; eauto using rt_refl.
            s; i; des; eauto using esim_progress_mon.
      - specialize (EST _ _ ST); des; ss; depdes ES; pclearbot; des
        ; [|by destruct evt|by inv SU RF].
        pfold; econs; [|edone].
        eapply beh_evt; eauto.
        right; eapply CIH; try apply CNF; eauto.
        by eapply esim_ext_impl_esim; eauto.
    }
    {
      destruct CNF.
      eapply RKa in RETa; eauto.
      eapply RKb in RETb; eauto.
      depdes MAT; try by exfalso; eauto using halted_step.
      by pfold; econs; eauto.
    }
  }
  {
    i; destruct SIM; eauto.
    {
      specialize (EST _ _ TAU); des
      ; depdes ES; pclearbot; [|by depdes SU RF VR].
      punfold ES.
      destruct ES.
      exploit ESIM; eauto.
      clear ESIM; intro ES; guardH ST; des; unguard.
      cut (behave r mbt obs).
      { i; eapply behave_taus; eauto.
        by des; subst; eauto using rt_trans, rt_step. }
      eapply IHTAU; eauto.
    }
    {
      destruct CNF.
      eapply RKa in RETa; eauto.
      eapply RKb in RETb; eauto.
      depdes MAT; by exfalso; eauto using halted_step.
    }
  }
Qed.


Lemma adequacy_msim: forall
    oenv ty (MN: lookup MAIN oenv = Some (FTarrow Tunit ty))
    modt mods (SIM: msim ts_gw nil oenv modt mods)
    
    (mt: mach tgt_lsi) (LDt: tgt_load modt = Some mt)
    (ms: mach src_lsi) (LDs: src_load mods = Some ms),
  behave bot2 mt <1= behave bot2 ms.
Proof.
  i; rrdes SIM.
  hexploit SIM; eauto; []; clear SIM; i; des.
  unfold tgt_load in *; destruct (load _ _ _) as [[gh lh]] eqn:E; des.
  i; hexploit (load_valid modt  0 nil); [by eauto|].
  intro X; dup X; split_pmvalid X.
  unfold src_load in LDs.

  set (cigt := {| cpc := Nothing; crf := Just (mrf mt); cst := mst mt;
                  chp := gh |}).
  set (cilt := {| cpc := Nothing; crf := Nothing; cst := pmemp;
                  chp := lh |}).
  set (cigs := (pmemp, Just (snd (fst ms)), Nothing) : conf src_lsi).
  set (cils := cemp : conf src_lsi).

  assert (CLDt: cload tgt_lsi modt 0 nil cigt cilt).
  { unfold cigt, cilt; destruct (lookup _ _) in LDt; vauto; s.
    clear X0; unfold tgt_cload; rewrite E; s.
    rr; esplits; s; eauto; tmext_simpl; auto.
    - i; apply pmread_some; auto.
    - instantiate (1 := pmemp); noemp; eauto.
    - split; [vauto|intro RD; des; apply pmread_empD in RD; auto].
    - noemp; auto.
  }

  assert (CLDs: cload src_lsi mods tt nil cigs cils).
  { unfold cigs, cils; destruct (lookup _ _) in LDs; vauto; s.
    exists nil; rewrite app_nil_l app_nil_r; esplits; eauto.
  }

  hexploit H; eauto; []; clear H; i; des; clear CLDt CLDs.

  specialize (SIM _ _ MN); des.

  rrdes Ea; simpl map in Ea; unfold label in Ea.
  rewrite E in Ea; rewrite Ea in LDt; clarify.
  rrdes Eb; unfold label in Eb; rewrite Eb in LDs; clarify.

  exploit SIM; eauto using gkmin_bot_gknow; [|by vauto|].
  { rr; eauto using gkmin_bot_gknow. }

  hexploit INIg; eauto using gkmin_bot_gknow; []; clear INIg; intro INIg.
  hexploit INIl; eauto using gkmin_bot_gknow; []; clear INIl; intro INIl.

  dup INIg; rrdes INIg; subst; rrdes WHX.
  hexploit tgt_confpred_shape; eauto; []; i; des; subst.
  hexploit src_confpred_shape; eauto; []; i; des; inst; subst.
  hexploit ts_refconfrel_shape; eauto; []; i; des; subst.
  s in HPA; s in HPB; noemp; clarify.
  rrdes CPa; s in PC; s in RF; s in ST; s in HP; clean.
  rrdes CPb; clean.

  clear SIM; i; invc x1 VR IN.

  rrdes CA CB; invc SVQ.
  revert PR; setmrf rf; fold rf in H2; i.
  hexploit (CS (rf r_ret) nil); eauto.
  { eapply call_app; eauto.
    - econs; eauto; rewrite H2; eauto.
      unfold rf; s; tmext_simpl; eauto.
    - instantiate (2 := Vunit); instantiate (2 := vb).
      inst_pairs; rr; eauto.
    - rewrite H2; unfold rf; tmext_simpl.
      econs 1; rr; splits; eauto; rewrite -H; eauto.
    - unfold rf; tmext_simpl.
      econs; eapply vclgen_unit; rr; split; eauto.
  }

  assert (refha = pmemp).
  { apply ts_refconfrel_emp in WHY; [|by i; eapply REF; rr; eauto].
    des; invc WHY.
  }
  subst refha; noemp.
  
  
  
  
  
  
  

  i; eapply behave_taus; [|by s; inst_pairs; econs 1; econs 1; eauto]; rewrite H.
  eapply (@adequacy_esim tgt_lsi src_lsi) with (w := wlift lw) (si := si) (s := si)
    (kia := rf r_ret) (kib := nil) (cfa := @cemp tgt_lsi) (cfb := @cemp src_lsi)
  ; eauto using gkmin_bot_gknow, extra_emp.
  - i; apply gkmin_unfold in G; des; vauto.
  - unfold rf; tmext_simpl.
    i; rrdes RET; invc RET; rrdes MA.
    destruct ma, ha; s in PC; destruct cpc in PC; vauto.
  - i; rrdes RET; invc RET; rrdes MB.
    destruct hb as [[]]; s in Ec; destruct e2; vauto.
    destruct vb0; vauto.
  - econs; first 3 last.
    + instantiate (1 := (_, Just _, Nothing)); s; noemp; rr; esplits; eauto.
    + rr; esplits; eauto; unfold cils; noemp; auto.
    + split; rr; s; eauto.
    + s; noemp; rr; esplits; s; vauto.
      * apply HP in CA; s in CA; dup CA.
        apply pmread_valid, pmvalid_some in CA0; des; subst.
        apply pmread_some in CA; rewrite CA; auto.
      * apply pmfinite_plus; eauto using load_gfinite, load_lfinite.
Qed.
